from abc import ABC, abstractmethod, abstractclassmethod
import urllib.parse
import typing
import contextlib

from ..artefacts import Artefact, File, Directory

class AbstractManager(ABC):

    @abstractmethod
    def __repr__(self):
        pass

    @abstractmethod
    def _abspath(self, managerPath: str) -> str:
        """ Return the absolute path on the backend provider from the standardised manager path.

        Args:
            managerPath: The manager relative path which is to be converted to an absolute path

        Returns:
            str: The manager absolute path

        Examples:
            For the filesystem, this will be the full absolute path to the object. For s3 this is the key of the object.

            >>> stow.connect(manager='FS', path='/home/ubuntu')._abspath('/hello/there')
            '/home/ubuntu/hello/there'
            >>> stow.connect(manager='s3', bucket='bucket-example')._abspath('/hello/there')
            'hello/there'

        """
        pass


    @abstractmethod
    def _identifyPath(self, managerPath: str) -> typing.Union[Artefact, None]:
        """ For the path given, create an `Artefact` for the object at the location on the manager but do not add it
        into the manager. If no object exists - return None

        Args:
            abspath: The path for artefact on disk

        Returns:
            typing.Union[Artefact, None]: The artefact object that represents the item on disk or None if nothing exists
        """
        pass

    @abstractmethod
    def _get(self, source: Artefact, destination: str):
        """ Fetch the artefact and downloads its data to the local destination path provided

        The existence of the file to collect has already been checked so this function can be written to assume its
        existence

        Args:
            source: The source object and context that is to be downloaded
            destination: The local path to where the source is to be written
        """
        pass

    @abstractmethod
    def _getBytes(self, source: Artefact) -> bytes:
        """ Fetch the file artefact contents directly. This is to avoid having to write the contents of files to discs
        for some of the other operations.

        The existence of the file to collect has already been checked so this function can be written to assume its
        existence

        Args:
            source: The source object and context that is to be downloaded

        Returns:
            bytes: The bytes content of the disk
        """
        pass

    @abstractmethod
    def _put(self, source: str, destination: str):
        """ Put the local filesystem object onto the underlying manager implementation using the absolute paths given.

        To avoid user error - an artefact cannot be placed onto a Directory unless an overwrite toggle has been passed
        which is False by default. This should protect them from accidentally deleting a directory.

        In the event that they want to do so - the deletion of the directory will be handled before operating this
        function. Therefore there is no need to check/protect against it. (famous last words)

        Args:
            source: A local absolute path to an artefact (File or Directory)
            destination: A manager abspath path for the artefact
        """
        pass

    @abstractmethod
    def _putBytes(self, fileBytes: bytes, destination: str):
        """ Put the bytes of a file object onto the underlying manager implementation using the absolute path given.

        This function allows processes to avoid writing files to disc for speedier transfers.

        If it's not possible to transmit bytes - I'd recommend writing the bytes to a tempfile and then operating the
        put method.

        Args:
            fileBytes (bytes): files bytes
            destinationAbsPath (str): Remote absolute path
        """
        pass

    @abstractmethod
    def _cp(self, source: Artefact, destination: str):
        """ Method for copying an artefact local to the manager to another location on the manager. Implementation
        would avoid having to download data from a manager to re-upload that data.

        If there isn't a method of duplicating the data on the manager, you can call
            self._put(self._abspath(source.path), destination)

        Which will mean the behaviour defaults to the put action.

        Args:
            source: the manager local source artefact
            destination: a manager abspath path for destination
        """
        pass

    @abstractmethod
    def _mv(self, source: Artefact, destination: str):
        """ Method for moving an artefact local to the manager to another location on the manager. Implementation
        would avoid having to download data from a manager to re-upload that data.

        If there isn't a method of duplicating the data on the manager, you can call
            self._put(self._abspath(source.path), destination)
            self._rm(self._abspath(source.path))

        Which will mean the behaviour defaults to the put action and then a delete of the original file. Achieving the
        same goal.

        Args:
            source: the manager local source file
            destination: a manager abspath path for destination
        """
        pass

    @abstractmethod
    def _ls(self, directory: str):
        """ List all artefacts that are present at the directory objects location and add them into the manager.

        Args:
            managerPath: the manager path to the directory whose content is to be indexed
        """
        pass

    @abstractmethod
    def _rm(self, artefact: Artefact):
        """ Delete the underlying artefact data on the manager.

        To avoid possible user error in deleting directories, the user must have already indicated that they want to
        delete everything

        Args:
            artefact: The artefact on the manager to be deleted
        """
        pass

    @abstractclassmethod
    def _signatureFromURL(cls, url: urllib.parse.ParseResult):
        """ Create the signature that can be passed to the init of the manager to create a new instance using the
        information passed via the url ParseResult object that will have been created via the stateless interface

        Args:
            url: The result of passing the stateless path through urllib.parse.urlparse

        Returns:
            Manager: A manager of this type loaded with information from the url
            Relpath: The manager relative path for the artefact that may have been referenced

        Raises:
            Error: Errors due to missing information and so on
        """
        pass

    @contextlib.contextmanager
    @abstractmethod
    def localise(self, artefact: typing.Union[Artefact, str]) -> str:
        """ Localise an artefact, ensure that there is an absolute path that reaches this artefact. For local artefacts
        this will be the direct abspath. Remote managers will get the artefact, and pass back the path to this local
        version.

        A path is still returned even if the artefact doesn't exist. It will be the responsibility of the calling method
        to handle what is localised.

        Args:
            artefact: The artefact path or artefact object

        Yields:
            str: The abspath for the artefact
        """
        pass

    @abstractmethod
    def toConfig(self) -> dict:
        """ Generate a config which can be unpacked into the `connect` interface to initialise this manager. To be
        used to seralise and de-seralise a manager object.

        NOTE Defaulted values or environment variables are not guaranteed to be saved

        Returns:
            dict: A dictionary of the kwargs of the init of the manager
        """
        pass


import paramiko.client
import paramiko.sftp_client
import paramiko.ssh_exception

import os
import io
import pytz
import stat
import typing
import datetime
import urllib.parse
from functools import wraps

from ..artefacts import Artefact, File, Directory
from ..manager import Manager, RemoteManager
from .. import exceptions

def _ensureConnection(function: typing.Callable) -> typing.Callable:
    """ Instance method decorator to ensure that a sftp connection exists for manager calls. In the event that it
    has timed out, a new connection will be attempted.

    Args:
        function: The function to wrap

    Returns:
        Callable: A new function ensuring sftp connection
    """

    @wraps(function)
    def connect(self, *args, **kwargs):

        try:
            # Attempt to perform action
            return function(self, *args, **kwargs)

        except paramiko.ssh_exception.NoValidConnectionsError:
            # Connection has dropped - reconnect and attempt function again.
            # Do not handle subsequent connection issues

            self._connect()
            return function(self, *args, **kwargs)

    return connect

class SSH(RemoteManager):
    """ Manage a filesystem hosted on a remote machine using the SSH protocol

    Args:
        hostname: The hostname of the computer to manage
        port: The port where ssh connections are being handled
        username: The username of the user on the remote machine that the manager will use
        password: The password for the user
        privateKey: The private key used for ssh authentication
        privateKeyFilePath: A path to the private key for ssh authentication
        sshConfigs: Additional ssh config files for configurations lookup. Extending open ssh config locations.
            Order of Read will be: passed configs, user config, global config.
    """

    # Construct the base configs paths - only store them if they exists in the environment
    BASE_CONFIGS = [
        x
        for x in [
            os.path.expanduser(os.path.join('~', '.ssh', 'config')),
            os.path.expandvars(os.path.join("$ProgramData", 'ssh', 'ssh_config')) if os.name == 'nt' else "/etc/ssh/ssh_config"
        ]
        if os.path.exists(x)
    ]

    def __init__(
        self,
        hostname: str,
        root: str = None,
        port: int = 22,
        username: str = None,
        password: str = None,
        privateKey: str = None,
        privateKeyFilePath: str = None,
        autoAddMissingHost: bool = True,
        timeout: float = 30,
        sshConfigs: typing.Iterable[str] = None
        ):
        super().__init__()

        # Define the ssh and sftp client objects
        self._sshClient = paramiko.client.SSHClient()
        self._ftpClient: paramiko.sftp_client.SFTPClient = None

        # Configure the ssh connection parameters - Determine whether the client can accept an unknown host
        self._autoAddMissingHost = autoAddMissingHost
        if autoAddMissingHost:
            self._sshClient.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        # Load in the configuration files and check for the hostname
        self._sshConfigs = (sshConfigs or [])
        for path in self._sshConfigs + self.BASE_CONFIGS:

            sshConfig = paramiko.SSHConfig()
            sshConfig.parse(open(path))

            for configHostname in sshConfig.get_hostnames():
                if hostname == configHostname or hostname == configHostname.lower():
                    # A configuration has been found

                    # Fetch the config values
                    config = sshConfig.lookup(configHostname)

                    # Fetch the configuration values if they haven't been overwritten by the interface
                    hostname = config['hostname']
                    port = (port or config.get('port'))
                    username = (username or config.get('user'))
                    password = (password or config.get('password'))
                    privateKeyFilePath = (privateKeyFilePath or config.get('identityfile', [None])[0])
                    timeout = (timeout or config.get('timeout'))

        # Save connection information
        self._hostname = hostname
        self._port = port
        self._username = username
        self._password = password
        self._privateKey = privateKey
        self._privateKeyFilePath = privateKeyFilePath
        self._autoAddMissingHost = autoAddMissingHost
        self._timeout = timeout

        # Save remote parameters
        self._root = root  # Root address for the user
        self._home = f'/home/{username}'

        # Assuming that manager is initialised jit for operation - open connection
        self._connect()

    def _connect(self):
        # Connect to the remove machine
        self._sshClient.connect(
            hostname=self._hostname,
            port=self._port,
            username=self._username,
            password=self._password,
            pkey=self._privateKey,
            key_filename=self._privateKeyFilePath,
            timeout=self._timeout
        )

        self._ftpClient = self._sshClient.open_sftp()

    def __repr__(self):
        pass

    def _abspath(self, managerPath: str) -> str:

        if self._root:
            # A root path has been set - everything must be done relative to this path

            if managerPath == '/':
                # Root path has been selected
                return self._root

            # Join the path onto root
            return self.join(self._root, managerPath, joinAbsolutes=True, separator='/')

        else:
            # no root path given - interact with path like normal FS (assume cwd is home)
            if os.path.isabs(managerPath):
                return managerPath

            else:
                return self.join(self._home, managerPath, separator='/')

    def _statsToArtefact(self, artefactStat, artefactPath):

        # Determine whether the artefact is a file or directory
        isDirectory = stat.S_ISDIR(artefactStat.st_mode)

        # # Created time
        # createdTime = datetime.datetime.utcfromtimestamp(stats.st_ctime)
        # createdTime = pytz.UTC.localize(createdTime)

        # Modified time
        modifiedTime = datetime.datetime.utcfromtimestamp(artefactStat.st_mtime)
        modifiedTime = pytz.UTC.localize(modifiedTime)

        # Access time
        accessedTime = datetime.datetime.utcfromtimestamp(artefactStat.st_atime)
        accessedTime = pytz.UTC.localize(accessedTime)

        if isDirectory:
            return Directory(
                self,
                artefactPath,
                modifiedTime=modifiedTime,
                accessedTime=accessedTime
            )

        else:
            return File(
                self,
                artefactPath,
                size=artefactStat.st_size,
                modifiedTime=modifiedTime,
                accessedTime=accessedTime
            )

    @_ensureConnection
    def _identifyPath(self, managerPath: str) -> typing.Union[str, Artefact]:
        """ Given a manager relative path, identify what object (if any) is at that location and store its information
        as a result. Should call _addArtefact in the call stack for the returned item to ensure that things are updated

        Args:
            relpath: Manager relative path

        Returns:
            typing.Union[str, Artefact]: return None, "file", "directory" for nothing exists, file exists, directory
                respectively
        """

        try:
            stats = self._ftpClient.lstat(self._abspath(managerPath))
        except FileNotFoundError:
            return None

        return self._statsToArtefact(stats, artefactPath=managerPath)

    @_ensureConnection
    def _ensureDestination(self, path: str) -> None:
        """ Ensure that the path is a directory - if no artefact is found at location create it, and make any parent
        directories that are needed for the path to be valid

        Args:
            path: The remote fs path to a location where a directory is to exist
        """

        subpath = path
        parts = []
        while True:
            try:
                # Try and get the path stat - File not found error if it doesn't exist
                artefact = self._statsToArtefact(self._ftpClient.stat(subpath), subpath)

                # Artefact will need to be a directory
                if not isinstance(artefact, Directory):
                    raise exceptions.ArtefactTypeError(f'Cannot ensure directory path {path} as subpath {subpath} is a file')

                # The path exists - make dir for the parts not yet created
                for part in parts:
                    subpath = self.join(subpath, part, separator='/')
                    self._ftpClient.mkdir(subpath)

                # All directories created
                break
            except FileNotFoundError:
                subpath, dirname = os.path.split(subpath)
                parts.insert(0, dirname)

    def _recursiveGetDirectory(self, path, destination):

        # Create the directory at the desintation
        os.mkdir(destination)

        # Get the files at this level
        artefacts = self._ftpClient.listdir_attr(path)

        for artefact in artefacts:
            # Construct the destination path
            sourceFilepath = self.join(path, artefact.filename)
            artefactDestinationPath = os.path.join(destination, artefact.filename)

            if stat.S_ISDIR(artefact.st_mode):
                # The artefact is a directory - recursively create it and its children
                self._recursiveGetDirectory(sourceFilepath, artefactDestinationPath)

            else:
                # The item is a file - pull and place the file in the newly created directory
                self._ftpClient.get(sourceFilepath, artefactDestinationPath)


    @_ensureConnection
    def _get(self, source: Artefact, destination: str):

        if isinstance(source, Directory):
            self._recursiveGetDirectory(self._abspath(source.path), destination)

        else:
            # Fetch the file item
            self._ftpClient.get(self._abspath(source.path), destination)

    @_ensureConnection
    def _getBytes(self, source: Artefact) -> bytes:
        with self._ftpClient.open(self._abspath(source.path), 'rb') as handle:
            return handle.read()

    @_ensureConnection
    def _put(self, source: str, destination: str):

        destinationAbs = self._abspath(destination)

        if os.path.isdir(source):

            # Create the target location
            self._ensureDestination(destinationAbs)

            sourcePathLength = len(source) + 1
            for root, dirs, files in os.walk(source):

                # Create the path of the destination
                dRoot = self.join(destinationAbs, root[sourcePathLength:], separator='/')

                # Make the directory
                for dirname in dirs:
                    self._ftpClient.mkdir(self.join(dRoot, dirname, separator='/'))

                # For each file at this point - construct their local absolute path and their relative remote path
                for file in files:
                    self._ftpClient.put(
                        os.path.join(root, file),
                        self.join(dRoot, file, separator='/')
                    )

            # Return the empty directory object
            return self._statsToArtefact(self._ftpClient.stat(destinationAbs), destination)

        else:
            # Ensure the directory the file is meant to exist in and then put the file into the location
            self._ensureDestination(os.path.dirname(destinationAbs))
            return self._statsToArtefact(self._ftpClient.put(source,destinationAbs), destination)

    @_ensureConnection
    def _putBytes(self, fileBytes: bytes, destination: str):
        absDestination = self._abspath(destination)
        self._ensureDestination(os.path.dirname(absDestination))

        stats = self._ftpClient.putfo(io.BytesIO(fileBytes), absDestination, confirm=True)

        return self._statsToArtefact(stats, destination)

    @_ensureConnection
    def _cp(self, source: Artefact, destination: str):
        sourcePath = self._abspath(source.path)
        destination = self._abspath(destination)

        if isinstance(source, Directory):
            self._sshClient.exec_command(f'cp -R {sourcePath} {destination}')

        else:
            self._sshClient.exec_command(f'cp {sourcePath} {destination}')

    @_ensureConnection
    def _mv(self, source: Artefact, destination: str):
        sourcePath = self._abspath(source.path)
        destination = self._abspath(destination)

        self._sshClient.exec_command(f'mv {sourcePath} {destination}')

    @_ensureConnection
    def _rm(self, artefact: Artefact):
        artefactPath = self._abspath(artefact.path)
        if isinstance(artefact, Directory):
            self._sshClient.exec_command(f'rm -rf {artefactPath}')

        else:
            self._sshClient.exec_command(f'rm -f {artefactPath}')

    @_ensureConnection
    def _ls(self, managerPath: str) -> Directory:

        absManagerPath = self._abspath(managerPath)

        for stats in self._ftpClient.listdir_attr(absManagerPath):

            artefactPath = self.join(managerPath, stats.filename, separator='/')

            self._addArtefact(self._statsToArtefact(stats, artefactPath))

    @classmethod
    def _signatureFromURL(cls, url: urllib.parse.ParseResult):

        # Extract the query data passed into
        queryData = urllib.parse.parse_qs(url.query)

        signature = {
            "hostname": url.hostname,
            "port": (url.port or 22),
            "username": url.username,
            "password": url.password,
            "privateKey": queryData.get("privateKey", [None])[0],
            "privateKeyFilePath": queryData.get("privateKeyFilePath", [None])[0],
            "autoAddMissingHost": queryData.get("autoAddMissingHost", [True])[0],
            "timeout": queryData.get("timeout", 30),
            "sshConfigs": queryData.get("sshConfig"),
        }

        return signature, (url.path or '/')

    def toConfig(self):

        return {
            "manager": "ssh",
            "hostname": self._hostname,
            "port": self._port,
            "username": self._username,
            "password": self._password,
            "privateKey": self._privateKey,
            "privateKeyFilePath": self._privateKeyFilePath,
            "autoAddMissingHost": self._autoAddMissingHost,
            "timeout": self._timeout,
            "sshConfigs": self._sshConfigs
        }


""" Functions for the stow package to provided a stateless method of interacting with various manager artefacts.
"""

import os
import typing
import urllib.parse
import contextlib
from functools import wraps

from .artefacts import Artefact
from .manager import Manager, LocalManager
from . import utils

def _getManager(artefact) -> typing.Tuple[Manager, str]:

    if isinstance(artefact, Artefact):
        manager = artefact.manager
        relpath = artefact.path

    elif isinstance(artefact, str):
        return utils.parseURL(artefact)

    else:
        raise TypeError("Artefact reference must be either `stow.Artefact` or string not type {}".format(type(artefact)))

    return manager, relpath

@wraps(utils.find)
def find(*args, **kwargs) -> typing.Type[Manager]:
    return utils.find(*args, **kwargs)

@wraps(utils.connect)
def connect(*args, **kwargs) -> Manager:
    return utils.connect(*args, **kwargs)

@wraps(utils.parseURL)
def parseURL(*args, **kwargs) -> utils.ParsedURL:
    return utils.parseURL(*args, **kwargs)

def artefact(stowPath: str) -> Artefact:
    """ Fetch an artefact object for the given path

    Params:
        stowPath: Manager relative path to artefact

    Returns:
        Arefact: The artefact object

    Raises:
        ArtefactNotFound: In the event that no artefact exists at the location given
    """
    manager, relpath = _getManager(stowPath)
    return manager[relpath]

@wraps(Manager.abspath)
def abspath(*args, **kwargs) -> str:
    return Manager.abspath(*args, **kwargs)

@wraps(Manager.basename)
def basename(*args, **kwargs) -> str:
    return Manager.basename(*args, **kwargs)

@wraps(Manager.name)
def name(*args, **kwargs) -> str:
    return Manager.name(*args, **kwargs)

@wraps(Manager.extension)
def extension(*args, **kwargs) -> str:
    return Manager.extension(*args, **kwargs)

@wraps(Manager.commonpath)
def commonpath(*args, **kwargs) -> str:
    return Manager.commonpath(*args, **kwargs)

@wraps(Manager.commonprefix)
def commonprefix(*args, **kwargs) -> str:
    return Manager.commonprefix(*args, **kwargs)

@wraps(Manager.dirname)
def dirname(*args, **kwargs) -> str:
    return Manager.dirname(*args, **kwargs)

@wraps(Manager.expanduser)
def expanduser(*args, **kwargs) -> str:
    return Manager.expanduser(*args, **kwargs)

@wraps(Manager.expandvars)
def expandvars(*args, **kwargs) -> str:
    return Manager.expandvars(*args, **kwargs)

@wraps(Manager.isabs)
def isabs(*args, **kwargs) -> str:
    return Manager.isabs(*args, **kwargs)

@wraps(Manager.join)
def join(*args, **kwargs) -> str:
    return Manager.join(*args, **kwargs)

@wraps(Manager.normcase)
def normcase(*args, **kwargs) -> str:
    return Manager.normcase(*args, **kwargs)

@wraps(Manager.normpath)
def normpath(*args, **kwargs) -> str:
    return Manager.normpath(*args, **kwargs)

@wraps(Manager.realpath)
def realpath(*args, **kwargs) -> str:
    return Manager.realpath(*args, **kwargs)

@wraps(Manager.relpath)
def relpath(*args, **kwargs) -> str:
    return Manager.relpath(*args, **kwargs)

@wraps(Manager.samefile)
def samefile(*args, **kwargs) -> str:
    return Manager.samefile(*args, **kwargs)

@wraps(Manager.sameopenfile)
def sameopenfile(*args, **kwargs) -> str:
    return Manager.sameopenfile(*args, **kwargs)

@wraps(Manager.samestat)
def samestat(*args, **kwargs) -> str:
    return Manager.samestat(*args, **kwargs)

@wraps(Manager.split)
def split(*args, **kwargs) -> str:
    return Manager.split(*args, **kwargs)

@wraps(Manager.splitdrive)
def splitdrive(*args, **kwargs) -> str:
    return Manager.splitdrive(*args, **kwargs)

@wraps(Manager.splitext)
def splitext(*args, **kwargs) -> str:
    return Manager.splitext(*args, **kwargs)

@wraps(Manager.md5)
def md5(*args, **kwargs) -> str:
    return Manager.md5(*args, **kwargs)


# Instance wrapped

@wraps(Manager.isfile)
def isfile(artefact: typing.Union[Artefact, str]) -> bool:
    manager, relpath = _getManager(artefact)
    return manager.isfile(relpath)

@wraps(Manager.isdir)
def isdir(artefact: typing.Union[Artefact, str]) -> bool:
    manager, relpath = _getManager(artefact)
    return manager.isdir(relpath)

@wraps(Manager.islink)
def islink(artefact: typing.Union[Artefact, str]) -> bool:
    manager, relpath = _getManager(artefact)
    return manager.islink(relpath)

@wraps(Manager.ismount)
def ismount(artefact: typing.Union[Artefact, str]) -> bool:
    manager, relpath = _getManager(artefact)
    return manager.ismount(relpath)

@wraps(Manager.getctime)
def getctime(artefact: typing.Union[Artefact, str]) -> bool:
    manager, relpath = _getManager(artefact)
    return manager.getctime(relpath)

@wraps(Manager.getmtime)
def getmtime(artefact: typing.Union[Artefact, str]) -> bool:
    manager, relpath = _getManager(artefact)
    return manager.getmtime(relpath)

@wraps(Manager.getatime)
def getatime(artefact: typing.Union[Artefact, str]) -> bool:
    manager, relpath = _getManager(artefact)
    return manager.getatime(relpath)

@wraps(Manager.exists)
def exists(artefact: typing.Union[Artefact, str]) -> bool:
    manager, relpath = _getManager(artefact)
    return manager.exists(relpath)

@wraps(Manager.lexists)
def lexists(artefact: typing.Union[Artefact, str]) -> bool:
    manager, relpath = _getManager(artefact)
    return manager.lexists(relpath)

@wraps(Manager.touch)
def touch(artefact, *args, **kwargs):
    manager, relpath = _getManager(artefact)
    return manager.touch(relpath)

@wraps(Manager.mkdir)
def mkdir(artefact, *args, **kwargs):
    manager, relpath = _getManager(artefact)
    return manager.mkdir(relpath,*args, **kwargs)

@wraps(Manager.localise)
@contextlib.contextmanager
def localise(artefact, *args, **kwargs):
    manager, relpath = _getManager(artefact)
    with manager.localise(relpath, *args, **kwargs) as abspath:
        yield abspath

@wraps(Manager.open)
@contextlib.contextmanager
def open(artefact, mode = "r", *args, **kwargs):
    manager, relpath = _getManager(artefact)
    with manager.open(relpath, mode, *args, **kwargs) as handle:
        yield handle

@wraps(Manager.ls)
def ls(artefact = os.path.curdir, **kwargs):
    manager, relpath = _getManager(artefact)
    return manager.ls(manager.abspath(relpath), **kwargs)

@wraps(Manager.get)
def get(src_remote, dest_local, *args, **kwargs):
    manager, relpath = _getManager(src_remote)
    manager.get(relpath, dest_local, *args, **kwargs)

@wraps(Manager.put)
def put(src, dest, *args, **kwargs):

    # Get the destination manager for the artifact
    dest_manager, dest_relpath = _getManager(dest)

    # Put the source path into the manager
    return dest_manager.put(src, dest_relpath, *args, **kwargs)

@wraps(Manager.cp)
def cp(src, dest, *args, **kwargs):
    srcM, srcP = _getManager(src)
    destM, destP = _getManager(dest)
    destM.cp(srcP, destP, *args, **kwargs)

@wraps(Manager.mv)
def mv(src, dest, *args, **kwargs):
    srcM, srcP = _getManager(src)
    destM, destP = _getManager(dest)
    assert srcM is destM
    srcM.mv(srcP, destP, *args, **kwargs)

@wraps(Manager.sync)
def sync(src, dest, *args, **kwargs):
    srcM, srcP = _getManager(src)
    destM, destP = _getManager(dest)

    # Call sync on the destination manager
    destM.sync(srcM[srcP], destP, *args, **kwargs)

@wraps(Manager.rm)
def rm(artefact, *args, **kwargs):
    manger, relpath = _getManager(artefact)
    manger.rm(relpath, *args, **kwargs)

@property
def supports_unicode_filenames():
    """ True if arbitrary Unicode strings can be used as file names (within limitations imposed by the file system). """
    return os.path.supports_unicode_filenames


import os
import re
import io
import abc
import typing
import urllib
import shutil
import tempfile
import contextlib

from .abstract_methods import AbstractManager
from .class_methods import ClassMethodManager
from .reloader import ManagerSeralisable

from ..class_interfaces import ManagerInterface, LocalInterface, RemoteInterface
from ..artefacts import Artefact, File, Directory, SubFile, SubDirectory
from .. import utils
from .. import exceptions

import logging
log = logging.getLogger(__name__)

class Manager(AbstractManager, ClassMethodManager, ManagerInterface, ManagerSeralisable):
    """ Manager Abstract base class - expressed the interface of a Manager which governs a storage option and allows
    extraction and placement of files in that storage container

    """

    SUPPORTS_UNICODE_FILENAMES = os.path.supports_unicode_filenames

    _ROOT_PATH = "/"
    _PLACEHOLDER = "placeholder.ignore"
    _READONLYMODES = ["r", "rb"]

    _MULTI_SEP_REGEX = re.compile(r"(\\{2,})|(\/{2,})")
    _RELPATH_REGEX = re.compile(r"^([a-zA-Z0-9]+:)?([/\\\w\.!-_*'() ]*)$")

    def __init__(self):
        self._root = Directory(self, self._ROOT_PATH)
        self._paths = {self._ROOT_PATH: self._root}
        self._submanagers = {}

    def __contains__(self, artefact: typing.Union[Artefact, str]) -> bool:
        if isinstance(artefact, Artefact):
            return self is artefact.manager
        else:
            return self.exists(artefact)

    def __getitem__(self, path: str) -> Artefact:
        """ Fetch an artefact from the manager. In the event that it hasn't been cached, look it up on the underlying
        implementation and return a newly created object. If it doesn't exist raise an error

        Args:
            managerPath: The manager relative path to fine the artefact with

        Returns:
            artefact: The artefact at the provided location path

        Raises:
            ArtefactNotFound: In the event that the path does not exist
        """

        # Clean the path given - ensures its a manager path
        path = self._managerPath(path)

        # If the path exists within the manager - return it
        if path in self._paths:
            return self._paths[path]

        # Load the artefact and return it
        return self._loadArtefact(path)

    def _managerPath(self, path: str) -> str:

        if not path:
            return "/"

        # Expand any environment variables but no home path and do not make absolute relative to local
        path = self.normpath(self.expandvars(path))

        if os.name == 'nt':
            # Make Windows paths (if provided) unix like
            if path.find(":") != -1:
                # Remove the drive name
                path = path[path.index(":")+1:]

            path = path.replace("\\", "/")

        if not path.startswith('/'):
            path = '/' + path

        return path

    def _ensureDirectory(self, managerPath: str) -> Directory:
        """ Fetch the owning `container` for the manager relative path given. In the event that no `container` object
        exists for the path, create one and recursively find its owning `container` to add it to. The goal of this
        function is to traverse up the hierarchy and ensure all the directory objects exist, and when they do quickly
        return the container they are in

        Args:
            path (str): The manager relative path for an `Artefact`

        Returns:
            Directory: The owning directory container, which may have just been created

        Raises:
            ArtefactNotFound: If the path given doesn't exist in the manager
            ArtefactTypeError: The path exists but it isn't a directory object which is what is expected
        """

        if managerPath in self._paths:
            # The path points to an already established directory
            directory = self._paths[managerPath]
            if isinstance(directory, File):
                raise exceptions.ArtefactTypeError("Invalid path given {}. Path points to a file {}.".format(managerPath, directory))

            return directory

        if not managerPath:
            # No path given - root is being asked for
            return self._root

        # Create a directory at this location, add it to the data store and return it
        art = self._identifyPath(managerPath)
        if art is None:
            raise exceptions.ArtefactNotFound("No artefact found at location {}".format(managerPath))

        elif isinstance(art, File):
            raise exceptions.ArtefactTypeError("Invalid path given {}. Path points to a file {}.".format(managerPath, art))

        self._addArtefact(art)  # Link it with any owner + submanagers
        return art

    def _addArtefact(self, artefact: Artefact):
        """ Add an artefact object into the manager data structures - do not add if the object has already been added

        Args:
            artefact: The artefact object to be added

        Raises:
            ArtefactNotMember: in the event that the artefact that is trying to be added was not created by this manager
        """

        if artefact.manager is not self:
            raise exceptions.ArtefactNotMember("Artefact {} is not a member of {} and couldn't be added".format(artefact, self))

        if artefact.path in self._paths:
            # The artefact was already member of the manager - Updating the original artefact with new data
            self._paths[artefact.path]._update(artefact)

        else:
            # The artefact is new - ensure the parent directory and add the artefact into the manager

            # Get the directory for the artefact - add the artefact to that directories contents
            directory = self._ensureDirectory(self.dirname(artefact))
            directory._add(artefact)

            # Add the artefact into the manager store
            self._paths[artefact.path] = artefact

        if self._submanagers:
            # Ensure that the artefact has been added to any sub managers this artefact resides in

            for uri, manager in self._submanagers.items():
                if artefact.path.startswith(uri):
                    # The artefact exists within the sub manager - pass the parent object
                    manager._cascadeAddArtefact(artefact)

    def _loadArtefact(self, managerPath: str) -> Artefact:
        """ Identify the type of the artefact and then add it to the manager

        This allows you to specify the method of adding the artefact and doesn't limit you to adding a single artefact.
        It may be better to load an entire directory and then return the targeted artefact

        Args:
            managerPath: The manager relative path to the object to be loaded

        Returns:
            Artefact: the artefact created at that location

        Raises:
            ArtefactNotFound: If the path doesn't lead to an artefact
        """

        # Check to see if there is an artefact that exists on disk
        obj = self._identifyPath(managerPath)
        if obj is None:
            raise exceptions.ArtefactNotFound("Couldn't locate artefact {}".format(managerPath))

        # Add the created artefact to the manager
        self._addArtefact(obj)

        return obj

    def _findArtefact(self, source: typing.Union[Artefact, str]) -> Artefact:
        """ Find artefact even if it isn't managed by this manager. Use the public methods to create a manager for the
        incoming source object and get from it the artefact object
        """


        sourceObject, _ = self._artefactFormStandardise(source)
        if sourceObject is None:
            # The artefact wasn't given and the path doesn't lead to an artefact on the manager

            result = urllib.parse.urlparse(source)

            # Find the manager that is correct for the protocol
            if result.scheme and result.netloc:
                manager = utils.find(result.scheme)
                return manager._loadFromProtocol(result)[result.path]

            else:
                # Local manager - start it at the base of the file system
                manager = utils.connect("FS", path="/")
                return manager[self.abspath(source)]

        return sourceObject

    def _artefactFormStandardise(self, artefact: typing.Union[Artefact, str], require=False) -> typing.Tuple[Artefact, str]:
        """ Convert the incoming object which could be either an artefact or relative path into a standardised form for
        both such that functions can be easily convert and use what they require

        Args:
            artObj (typing.Union[Artefact, str]): Either the artefact object or it's relative path to be standardised
            require (str): Require that the object exists. when false return None for yet to be created objects but

        Returns:
            Artefact or None: the artefact object or None if it doesn't exists and require is False
            str: The relative path of the object/the passed value
        """
        if isinstance(artefact, Artefact):
            return artefact, artefact.path

        else:
            if require:
                # The artefact must be collected
                obj = self[artefact]
                return obj, obj.path

            # Clean and ensure the artefact path
            artefact = self._managerPath(artefact)

            if self.exists(artefact):
                # Check to see if the artefact exists - if it does then it is available
                return self._paths[artefact], artefact

            else:
                # The object doesn't exist on disk
                return None, artefact

    def _updateArtefactObjects(self, artefact: Artefact, identity: Artefact = None):
        """ Perform a update for the manager on the contents of a directory which has been editted on mass and whose
        content is likely inconsistent with the current state of the manager. Only previously known files are checked as
        new files are to be loaded JIT and can be added at that stage.

        Args:
            artobj (Directory): The directory to perform the refresh on
            identity (Artefact) = None: The new object artefact info
        """

        if isinstance(artefact, File):
            file = (identity or self._identifyPath(artefact.path))
            if not isinstance(file, File):
                # The object is no longer the same type - the original needs to be removed
                return self._delinkArtefactObjects(artefact)

            artefact._update(file)

        else:
            artefact: Directory

            # For the artefacts we know about - check their membership
            for subArtefact in list(artefact._contents):

                # Have the path checked
                check = self._identifyPath(subArtefact.path)

                # Update the artefact according to its state on disc
                if check is None or type(subArtefact) != type(check):
                    # The artefact has been deleted or the type of the artefact has changed - it needs to be delinked
                    self._delinkArtefactObjects(subArtefact)

                elif isinstance(subArtefact, File):
                    # Update the artefact with the informant as we've pulled it
                    subArtefact._update(check)

                else:
                    # The directory needs to be checked for issues
                    self._updateArtefactObjects(subArtefact)

            # Cannot be sure that all of the contents has been collected due to change
            artefact._collected = False

    def _moveArtefactObjects(self, srcObj: Artefact, destPath: str):
        """ Move a source an artefact to a new """

        if isinstance(srcObj, Directory):
            # Need to loop over directory contents and update their paths - their directory membership is fine

            for art in srcObj._ls(True):

                # Remove the artefact from its position in the manager
                del self._paths[art.path]

                # Update the object with it's new path
                art._path = self.join(destPath, srcObj.relpath(art), separator='/')

                # Update its membership
                self._paths[art.path] = art

        # Check whether the object has moved outside of the directory it was originally in
        if self.dirname(srcObj.path) != self.dirname(destPath):
            # Disconnect object with the directories that it exists in and add it to the destination location
            srcObj.directory._remove(srcObj)
            self._ensureDirectory(self.dirname(destPath))._add(srcObj)

        # Move the object in the manager state
        self._paths[destPath] = self._paths.pop(srcObj.path)

        # Update the artefacts info
        source_path = srcObj.path
        srcObj._path = destPath

        if self._submanagers:
            for uri, manager in self._submanagers.items():
                if srcObj.path.startswith(uri):
                    # The originating files have moved within the sub manager

                    if destPath.startswith(uri):
                        # The destination is within the sub-manager also
                        manager._cascadeMoveArtefactObjects(source_path, destPath)

                    else:
                        # The destination is outside the sub-manager - the subfiles need to be deleted
                        manager._cascadeDelinkArtefactObjects(srcObj)

    def _delinkArtefactObjects(self, artefact: Artefact):
        """ Unreference an artefact from the manager but do not check against/remove objects from the underlying
        implementation. This is to be used in conjunction with `_rm()` or to clean up artefacts that could have been
        affected as a side effect

        Args:
            artefact (Artefact): Manager artefact that is to be deleted
        """
        if isinstance(artefact, Directory):
            # Loop through cached directory contents and free those artefacts
            for childArtefact in artefact._ls(True):
                # NOTE directory only holds weakrefs to objects + the parent directory is being deleted so all
                # child objects should be deleted just fine.
                del self._paths[childArtefact.path]
                childArtefact._exists = False

        # Delete references to the object and set it's existence to false
        self[self.dirname(artefact.path)]._remove(artefact)
        del self._paths[artefact.path]

        # Get the artefact path before we make it existence False and cannot have access
        artefactPath = artefact.path

        # Make the artefact non existent
        artefact._exists = False

        if self._submanagers:
            for uri, manager in self._submanagers.items():
                if artefactPath.startswith(uri):
                    manager._cascadeDelinkArtefactObjects(artefactPath)

    def isfile(self, artefact: typing.Union[Artefact, str]) -> bool:
        """ Check if the artefact provided is a file

        Args:
            artefact: The artefact to be checked

        Returns:
            Bool: True or False in answer to the question
        """
        if isinstance(artefact, Artefact):
            return isinstance(artefact, File)

        return isinstance(self[artefact], File) if self.exists(artefact) else False

    def isdir(self, artefact: typing.Union[Artefact, str]) -> bool:
        """ Check if the artefact provided is a directory

        Args:
            artefact: The artefact to be checked

        Returns:
            Bool: True or False in answer to the question
        """


        if isinstance(artefact, Directory):
            return isinstance(artefact, File)

        return isinstance(self[artefact], Directory) if self.exists(artefact) else False

    def islink(self, artefact: typing.Union[Artefact, str]) -> bool:
        """ Check if the artefact provided is a link

        Will check for local managers but will default to False for remote managers

        Args:
            artefact: The artefact to be checked

        Returns:
            Bool: True or False in answer to the question
        """
        if isinstance(self, LocalManager):
            # We are local and we can use the local test
            return os.path.islink(self.abspath(artefact))

        else:
            log.warning("islink: Symbolic links are not supported - defaulting to False")
            return False

    def ismount(self, artefact: typing.Union[Artefact, str]) -> bool:
        """ Check if the artefact provided is a link

        Will check for local managers but will default to False for remote managers

        Args:
            artefact: The artefact to be checked

        Returns:
            Bool: True or False in answer to the question
        """
        if isinstance(self, LocalManager):
            # We are local and we can use the local test
            return os.path.ismount(self.abspath(artefact))

        else:
            log.warning("ismounts: mount point are not implemented on remote managers - returning False")
            return False

    def getctime(self, artefact: typing.Union[Artefact, str]) -> typing.Union[float, None]:
        """ Get the created time for the artefact as a UTC timestamp

        Args:
            artefact: the artefact whose creation datetime is to be returned

        Returns:
            timestamp: a float timestamp of creation time if manager holds such information else None

        Raises:
            ArtefactNotFound: If there is no artefact at the location
        """
        obj, _ = self._artefactFormStandardise(artefact, require=True)
        return obj.createdTime.timestamp()

    def getmtime(self, artefact: typing.Union[Artefact, str]) -> typing.Union[float, None]:
        """ Get the modified time for the artefact as a UTC timestamp

        Args:
            artefact: the artefact whose modified datetime is to be returned

        Returns:
            timestamp: a float timestamp of modified time if manager holds such information else None

        Raises:
            ArtefactNotFound: If there is no artefact at the location
        """
        obj, _ = self._artefactFormStandardise(artefact, require=True)
        return obj.modifiedTime.timestamp()

    def getatime(self, artefact: typing.Union[Artefact, str]) -> typing.Union[float, None]:
        """ Get the accessed time for the artefact as a UTC timestamp

        Args:
            artefact: the artefact whose accessed datetime is to be returned

        Returns:
            timestamp: a float timestamp of accessed time if manager holds such information else None

        Raises:
            ArtefactNotFound: If there is no artefact at the location
        """
        obj, _ = self._artefactFormStandardise(artefact, require=True)
        return obj.accessedTime.timestamp()

    def exists(self, artefact: typing.Union[Artefact, str]) -> bool:
        """ Return true if the given artefact is a member of the manager, or the path is correct for the manager and it
        leads to a File or Directory.

        Does not handle protocols

        Args:
            artefact: Artefact or path whose existence is to be checked

        Returns:
            bool: True if artefact exists else False
        """
        # Only split as we want to trigger reassessment of the underlying file
        obj, path = self._splitArtefactUnionForm(artefact)

        # Check to see if the artefact exists on disc still
        check = self._identifyPath(self._managerPath(path))

        # Check + update state if information about the object has changed
        if obj is not None:

            if check is not None:
                # Both objects exist - check that they have changed type and update if not

                if type(obj) != type(check):

                    # The obj has changed type
                    self._delinkArtefactObjects(obj)

                    # Add the artefact in
                    self._addArtefact(check)

                else:
                    # An update has occurred
                    if isinstance(obj, File):
                        obj._update(check)

            else:
                # The on disk representation has been removed
                self._delinkArtefactObjects(obj)

        elif check is not None:
            # A new artefact has been created
            self._addArtefact(check)

        # If not none then its a valid object - any updates will have taken place
        return check is not None

    def lexists(self, artefact: typing.Union[Artefact, str]) -> str:
        """ Return true if the given artefact is a member of the manager, or the path is correct for the manager and it
        leads to a File or Directory.

        Does not handle protocols

        Args:
            artefact: Artefact or path whose existence is to be checked

        Returns:
            bool: True if artefact exists else False
        """
        log.warning("lexists: Symbolic links are not supported - defaulting to exists")
        return self.exists(artefact)

    def get(self, source: typing.Union[Artefact, str], destination: str = None, overwrite: bool = False) -> Artefact:
        """ Get a remote artefact from the storage option and write it to the destination path given.

        Args:
            source (Artefact/str): The remote's file object or its path
            destination (str) = None: The local path for the artefact to be written to

        Returns:
            typing.Union[typing.Any, bytes]: Return user defined response for get if file written to destination else
                return bytes if no destination given
        """

        # Ensure the destination - Remove or raise issue for a local artefact at the location where the get is called
        if destination is not None:
            if os.path.exists(destination):
                if os.path.isdir(destination):
                    if overwrite:
                        shutil.rmtree(destination)

                    else:
                        raise exceptions.OperationNotPermitted(
                            "Cannot replace local directory ({}) unless overwrite argument is set to True".format(destination)
                        )

                else:
                    os.remove(destination)

            else:
                # Ensure the directory that this object exists with
                os.makedirs(self.dirname(destination), exist_ok=True)

        # Split into object and path - Ensure that the artefact to get is from this manager
        obj, path = self._artefactFormStandardise(source, require=True)
        if obj.manager is not self:
            raise exceptions.ArtefactNotMember("Provided artefact is not a member of the manager")

        # Fetch the object and place it at the location
        if destination is not None:
            return self._get(obj, destination)

        else:
            if not isinstance(obj, File):
                raise exceptions.ArtefactTypeError("Cannot get file bytes of {}".format(obj))

            return self._getBytes(obj)

    def put(
        self,
        source: typing.Union[Artefact, str, bytes],
        destination: typing.Union[Artefact, str],
        overwrite: bool = False,
        ) -> Artefact:
        """ Put a local artefact onto the remote at the location given.

        Args:
            src_local (str): The path to the local artefact that is to be put on the remote
            dest_remote (Artefact/str): A file object to overwrite or the relative path to a destination on the
                remote
            overwrite (bool) = False: Whether to accept the overwriting of a target destination when it is a directory
        """

        # Break the destination object apart
        destinationObj, destinationPath = self._artefactFormStandardise(destination)

        # Validated and prepare the destination
        if destinationObj is not None:

            # Check destination is member
            if destinationObj.manager is not self:
                raise exceptions.ArtefactNotMember("Destination artefact is not a member of the manager")

            # Delete the object to make space for new object
            if overwrite or isinstance(destinationObj, File) :
                # Remove the destination object
                self._rm(destinationObj)

            else:
                raise exceptions.OperationNotPermitted(
                    "Cannot put {} as destination is a directory, and overwrite has not been set to True"
                )

        # Process the source and put it onto the manager
        putArtefact = None
        if isinstance(source, bytes):
            # Source is file bytes - pass to manager implementation
            putArtefact = self._putBytes(source, destinationPath)

        else:
            # Source is a artefact in persisted storage
            if isinstance(source, str):
                # Turn the path into an artefact object even for external managers
                source = self._findArtefact(self.abspath(source))

            # Ensure that the artefact can be put by localising it
            with source.localise() as abspath:
                # Put the artefact into the destination
                putArtefact = self._put(abspath, destinationPath)

        # Post put cleanup - process the destination object
        if destinationObj is not None:
            # Validate whether the destination obj needs to be updated or removed as a consequence

            # Check where the type of the source is the same as the type of the destination object
            if isinstance(source, (bytes, File)) == isinstance(destinationObj, File):
                # The source placed is the same type as the established destination object - can update
                self._updateArtefactObjects(destinationObj, identity=putArtefact)

                # Return the original and updated destination object
                return destinationObj

            else:
                # Different object type now so we need to get rid of the original artefact and replace
                self._delinkArtefactObjects(destinationObj)

        # Create a new artefact object and return
        if putArtefact is not None:
            self._addArtefact(putArtefact)
            return putArtefact

        else:
            return self._loadArtefact(destinationPath)

    def cp(
        self,
        source: typing.Union[Artefact, str],
        destination: typing.Union[Artefact, str],
        overwrite: bool = False
        ) -> Artefact:
        """ Copy the artefacts at the source location to the provided destination location. Overwriting items at the
        destination.

        Args:
            source: source path or artefact
            destination: destination path or artefact
            overwrite: Whether to overwrite directories my move

        Returns:
            Artefact: The destination artefact object
        """

        # Ensure the destination - get the destination object
        destinationObj, destinationPath = self._artefactFormStandardise(destination)
        if destinationObj:

            # Only work on targets that are on the manager
            if destinationObj.manager is not self:
                raise exceptions.ArtefactNotMember(
                    "Cannot copy onto an artefact {} not within manager {}".format(destinationObj, self)
                )

            # There is an artefact at that location that will need to be removed - check if that is allowed
            if isinstance(destinationObj, Directory) and not overwrite:
                raise exceptions.OperationNotPermitted(
                    "Cannot copy artefact to location as destination is a directory {} and overwrite has not been toggled".format(
                        destinationObj
                    )
                )

            # Remove the original object
            self._rm(destinationObj.path)
            self._delinkArtefactObjects(destinationObj)

        # Look to see if the source artefact is in the manager - if so we can try to be more efficient
        sourceObject, sourcePath = self._artefactFormStandardise(source)
        if sourceObject is None or sourceObject.manager is not self:
            # The source isn't inside this manager - we must find it and use put. No speed up to be had in manager

            if sourceObject is None:
                # Find the original source object
                sourceObject = self._findArtefact(source)

            return self.put(sourceObject, destinationPath)

        # We must be an artefact on the box copying to another location on the box - destination is clear
        self._cp(sourceObject, destinationPath)
        return self[destinationPath]

    def mv(
        self,
        source: typing.Union[Artefact, str],
        destination: typing.Union[Artefact, str],
        overwrite: bool = False
        ) -> Artefact:
        """ Copy the artefacts at the source location to the provided destination location. Overwriting items at the
        destination.

        Args:
            source: source path or artefact
            destination: destination path or artefact
            overwrite: Whether to overwrite directories my move

        Returns:
            Artefact: The destination artefact object (source object updated if source was on manager originally)
        """

        # Ensure the destination - get the destination object
        destinationObj, destinationPath = self._artefactFormStandardise(destination)
        if destinationObj:

            # Only work on targets that are on the manager
            if destinationObj.manager is not self:
                raise exceptions.ArtefactNotMember(
                    "Cannot move onto an artefact {} not within manager {}".format(destinationObj, self)
                )

            # There is an artefact at that location that will need to be removed - check if that is allowed
            if isinstance(destinationObj, Directory) and not overwrite:
                raise exceptions.OperationNotPermitted(
                    "Cannot move artefact to location as destination is a directory {} and overwrite has not been toggled".format(
                        destinationObj
                    )
                )

            # Remove the original object
            self._rm(destinationObj.path)
            self._delinkArtefactObjects(destinationObj)

        # Look to see if the source artefact is in the manager - if so we can try to be more efficient
        sourceObject, _ = self._artefactFormStandardise(source)
        if sourceObject is None or sourceObject.manager is not self:
            # The source isn't inside this manager - we must find it and use put. No speed up to be had in manager

            if sourceObject is None:
                # Find the original source object
                sourceObject = self._findArtefact(source)

            artefact = self.put(sourceObject, destinationPath)
            sourceObject.manager.rm(sourceObject)  # Delete the original source as it has been moved
            return artefact

        # The source is on the manager and can be moved with the underlying infrastructure
        self._mv(sourceObject, destinationPath)
        self._moveArtefactObjects(sourceObject, destinationPath)
        return self[destinationPath]

    def rm(self, artefact: typing.Union[Artefact, str], recursive: bool = False) -> None:
        """ Remove an artefact from the manager using the artefact object or its relative path. If its a directory,
        remove it if it is empty, or all of its contents if recursive has been set to true.

        Args:
            artefact (typing.Union[Artefact, str]): the object which is to be deleted
            recursive (bool) = False: whether to accept the deletion of a directory which has contents
        """

        obj, _ = self._artefactFormStandardise(artefact, require=True)

        if obj is None or obj.manager is not self:
            raise exceptions.ArtefactNotMember("Artefact ({}) is not a member of the manager".format(artefact))

        if isinstance(obj, Directory) and len(obj) and not recursive:
            raise exceptions.OperationNotPermitted(
                "Cannot delete a container object that isn't empty - set recursive to True to proceed"
            )

        # Remove the artefact from the manager
        self._rm(obj)  # Remove the underlying data objects
        self._delinkArtefactObjects(obj)  # Remove references in the manager and set the objects._exist = False

    def ls(self, art: typing.Union[Directory, str] = '/', recursive: bool = False) -> typing.Set[Artefact]:
        """ List contents of the directory path/artefact given.

        Args:
            art (Directory/str): The Directory artefact or the relpath to the directory to be listed
            recursive (bool) = False: Return subdirectory contents as well

        Returns:
            {Artefact}: The artefact objects which are within the directory
        """

        # Convert the incoming artefact reference - require that the object exist and that it is a directory
        artobj, _ = self._artefactFormStandardise(art, require=True)
        if not isinstance(artobj, Directory):
            raise TypeError("Cannot perform ls action on File artefact: {}".format(artobj))

        # Perform JIT download of directory contents
        if not artobj._collected:
            self._ls(artobj.path)
            artobj._collected = True

        if recursive:

            # Iterate through contents and recursively add lower level artifacts
            contents = set()
            for art in artobj._contents:
                if isinstance(art, Directory): contents |= self.ls(art, recursive)
                contents.add(art)

            # Return all child content
            return contents

        return set(artobj._contents)

    def mkdir(self, path: str, ignoreExists: bool = True, overwrite: bool = False) -> Directory:
        """ Make a directory at the location of the path provided. By default - do nothing in the event that the
        location is already a directory object.

        Args:
            path (str): Relpath to the location where a directory is to be created
            ignoreExists (bool) = True: Whether to do nothing if a directory already exists
            overwrite (bool) = False: Whether to overwrite the directory with an empty directory

        Returns:
            Directory: The directory at the given location - it may have been created as per the call

        Raises:
            OperationNotPermitted: In the event that you try to overwrite a directory that already exists without
                passing the overwrite flag
        """

        if path in self:
            art = self[path]

            if isinstance(art, File):
                raise exceptions.OperationNotPermitted("Cannot make a directory as location {} is a file object".format(path))

            if ignoreExists and not overwrite:
                return art

        with tempfile.TemporaryDirectory() as directory:
            return self.put(directory, path, overwrite=overwrite)

    def touch(self, relpath: str) -> Artefact:
        return self.put(b'', relpath)

    def sync(self, source: typing.Union[Directory, str], destination: typing.Union[Directory, str], overwrite: bool = False, delete: bool = False) -> None:
        """ Put artefacts in the source location into the destination location if they have more recently been editted

        Args:
            source (Directory): source directory artefact
            destination (Directory): destination directory artefact on the manager
            delete: Togger the deletion of artefacts that are members of the destination which do not conflict with
                the source.

        Raises:
            ArtefactNotFound: In the event that the source directory doesn't exist

        """

        # Fetch the destination object - If None, nothing to sync so simply put the source
        destObj, destPath = self._artefactFormStandardise(destination)
        if destObj is None:
            return self.put(source, destPath)

        # Fetch the source object and require that it be an Artefact so we can check object states
        source = self._findArtefact(source)

        # Ensure that the two passed artefacts are directories
        if not (isinstance(source, Directory) and isinstance(destObj, Directory)):
            raise exceptions.ArtefactTypeError("Cannot Synchronise non directory objects {} -> {} - must sync directories".format(source, destination))

        # Get the mappings of source artefacts and destination objects
        sourceMapped = {
            source.relpath(artefact): artefact
            for artefact in source.ls(recursive=True)
            if isinstance(artefact, File)
        }

        destinationMapped = {
            destObj.relpath(artefact): artefact
            for artefact in destObj.ls(recursive=True)
        }

        # Iterate over all the files in the source
        for relpath, sourceArtefact in sourceMapped.items():

            # Look to see if there is a conflict
            if relpath not in destinationMapped:
                # The file doesn't conflict so we will push to destination
                self.put(sourceArtefact, self.join(destObj.path, relpath, separator='/'))

            else:
                # There is a conflict - lets compare local and destination
                destinationArtefact = destinationMapped.pop(relpath)

                # Don't perform sync
                if isinstance(destinationArtefact, Directory) and not overwrite:
                    raise exceptions.OperationNotPermitted(
                        "Cannot sync source file {} to destination is a directory {}, and operation not permitted".format(
                            sourceArtefact, destinationArtefact
                        )
                    )

                elif sourceArtefact.modifiedTime > destinationArtefact.modifiedTime:
                    # File is more up to date than destination
                    self.put(sourceArtefact, destinationArtefact)

        # Remove destination artefacts if delete is toggled
        if delete:
            # As updated artefacts were popped during their sync - any left File artefacts are to be deleted

            # Sort to ensure that nested files appear before their directories
            # This allows us to check to see if the directory is empty knowning that everything to be deleted from it
            # has been, as all nested artefacts will appear sooner and be removed before hand.
            for artefact in sorted(destinationMapped.values(), key=lambda x: len(x.path)):

                if isinstance(artefact, File) or (artefact.isEmpty()):
                    # Delete the file or remove the directory if it is now empty
                    self.rm(artefact)

    @contextlib.contextmanager
    def open(self, artefact: typing.Union[File, str], mode: str = "r", **kwargs) -> io.IOBase:
        """ Open a file and create a stream to that file. Expose interface of `open`

        Args:
            artefact: The object that represents the file (or path to the file) to be openned by this manager
            mode: The open method
            kwargs: kwargs to be passed to the interface of open

        Yields:
            io.IOBase: An IO object depending on the mode for interacting with the file
        """

        art, path = self._artefactFormStandardise(artefact)

        if art is None:
            if mode in self._READONLYMODES:
                raise FileNotFoundError('File does not exist in the manager')

        with self.localise(artefact) as abspath:
            with open(abspath, mode, **kwargs) as handle:
                yield handle

    def submanager(self, uri: str):
        """ Create a submanager at the given uri which shall behave like a conventional manager, however, its actions
        shall be relative to the given uri and shall update the main manager.

        If a manager exists at the uri specified already, then it is returned.

        Args:
            uri (str): The uri of the target location for the manager to be setup. If the uri does not exist, a
                directory shall be created. If it exists, the manager shall require it be a Directory object

        Returns:
            SubManager: A sub manager at the given uri

        Raises:
            ValueError: Raised if uri is top level directory
            ArtefactTypeError: if there exists an object at the location which isn't a directory
        """
        if uri == "/": raise ValueError("Cannot create a sub-manager at the top level of a manager")
        if uri in self._submanagers: return self._submanagers[uri]

        # Get or make the uri directory
        try:
            art = self[uri]
        except exceptions.ArtefactNotFound:
            art = self.mkdir(uri)

        # Ensure it is a directory and return + save the manager
        if isinstance(art, Directory):
            manager = SubManager(self, uri, art)
            self._submanagers[uri] = manager
            return manager
        else:
            raise exceptions.ArtefactTypeError("Cannot create a submanager with a file's path")

class SubManager(Manager):
    """ Created by a `Manager` instance to manage a section of the filesystem as if it were a fully fledged manager. The
    interface passes through to owning manager who executes the actions asked to the Sub Manager. Not to be instantiated
    directly or extended.

    Args:
        owner: The manager object this submanager belongs too
        path: The path in the manager the submanager exists
        rootDirectory: The owning managers root artefact object
    """

    def __init__(self, owner: Manager, path: str, rootDirectory: Directory):
        self._root = SubDirectory(self, self._ROOT_PATH, rootDirectory)
        self._path = path
        self._paths = {self._ROOT_PATH: self._root}
        self._submanagers = None

        self._owner = owner

    def __repr__(self):
        return '<SubManager of {} {}>'.format(self._owner, self._path)

    @classmethod
    def _relpath(cls, path: str, target: str):
        relpath = "/" + cls.relpath(path, target)
        if os.name == 'nt':
            relpath = relpath.replace("\\", "/")
        return relpath

    def _join(self, *paths):
        """ Join manager paths with this base manager path for full concrete manager path

        Args:
            *paths: The paths objects to join

        Returns:
            str: the concrete manager path
        """
        return self.join(self._path, *paths, joinAbsolutes=True, separator="/")

    def _cascadeAddArtefact(self, artefact: Artefact):
        # An artefact has either been updated or added to the manager that should also be present in this sub manager

        # Get the new artefact's path
        subpath = self._relpath(artefact.path, self._path)

        if subpath in self._paths:
            # The artefact is represented and out subartefact will already be updated as a result of the parent's actions
            return

        # The artefact is new - we will create the sub artefact objects and add them with the same function as our parents
        # NOTE as there cannot be any submanagers of this sub manager this is not recursively breaking
        if isinstance(artefact, File):
            subArtefact = SubFile(self, subpath, artefact)
        else:
            subArtefact = SubDirectory(self, subpath, artefact)

        super()._addArtefact(subArtefact)

    # Overload the update method - there is no cascade as child artefacts will pull from updated parents
    def _updateArtefactObjects(self, artefact: Artefact, identity: Artefact = None):
        return self._owner._updateArtefactObjects(artefact._concrete, identity=identity)

    # Overload the move method
    def _moveArtefactObjects(self, source: Artefact, destination: str):
        return self._owner._moveArtefactObjects(source._concrete, self._join(destination))

    def _cascadeMoveArtefactObjects(self, source: str, destination: str):
        # Run the parent move function on the subartefacts
        super()._moveArtefactObjects(self._paths[self._relpath(source, self._path)], self._relpath(destination, self._path))

    # Overload the delink method
    def _delinkArtefactObjects(self, artefact: Artefact):
        return self._owner._delinkArtefactObjects(artefact._concrete)
    def _cascadeDelinkArtefactObjects(self, artefact: str):
        return super()._delinkArtefactObjects(self._paths[self._relpath(artefact, self._path)])

    def _abspath(self, managerPath):
        return self._owner._abspath(self._join(managerPath))

    def _makeFile(self, managerPath: str) -> File:
        mainArt = self._owner._makeFile(self._join(managerPath))
        return SubFile(self, managerPath, mainArt)

    def _makeDirectory(self, managerPath: str) -> Directory:
        mainDirectory = self._owner._makeDirectory(self._join(managerPath))
        return SubDirectory(self, managerPath, mainDirectory)

    def _identifyPath(self, managerPath: str) -> typing.Union[Artefact, None]:
        mainCheck = self._owner._identifyPath(self._join(managerPath))

        if isinstance(mainCheck, File):
            return SubFile(self, managerPath, mainCheck)

        elif isinstance(mainCheck, Directory):
            return SubDirectory(self, managerPath, mainCheck)

        else:
            return mainCheck

    def _loadArtefact(self, managerPath: str) -> Artefact:
        self._owner._loadArtefact(self._join(managerPath))
        return self._paths[managerPath]

    def _addArtefact(self, artefact: Artefact):
        # Add the concrete artefact to the owning manager - add to local first to ensure that no new sub artefact is
        # created
        super()._addArtefact(artefact)
        self._owner._addArtefact(artefact._concrete)

    def _get(self, source: Artefact, destination: str):
        return self._owner._get(source._concrete, destination)

    def _getBytes(self, source: Artefact) -> bytes:
        return self._owner._getBytes(source._concrete)

    def _put(self, source: str, destination: str):
        self._owner._put(source, self._join(destination))

    def _putBytes(self, fileBytes: bytes, destination: str):
        self._owner._putBytes(fileBytes, self._join(destination))

    def _cp(self, source: Artefact, destination: str):
        return self._owner._cp(source._concrete, self._join(destination))

    def _mv(self, source: Artefact, destination: str):
        return self._owner._mv(source._concrete, self._join(destination))

    def _rm(self, artefact: Artefact):
        self._owner._rm(artefact._concrete)

    def _ls(self, directory: str):
        return self._owner._ls(self._join(directory))

    @contextlib.contextmanager
    def localise(self, artefact: typing.Union[Artefact, str]):
        with type(self._owner).localise(self, artefact) as abspath:
            yield abspath

    def submanager(self):
        raise NotImplementedError("A submanager cannot be created on a submanager")

    @classmethod
    def _signatureFromURL(cls, url: urllib.parse.ParseResult):
        raise NotImplementedError("Cannot load a submanager from a protocol string")

    def toConfig(self) -> dict:
        config = self._owner.toConfig()

        # Add this managers uri as submanager point
        config["submanager"] = self._path

        return config

class LocalManager(Manager, abc.ABC, LocalInterface):
    """ Abstract Base Class for managers that will be working with local artefacts.
    """

    @contextlib.contextmanager
    def localise(self, artefact):
        obj, path = self._artefactFormStandardise(artefact)
        exception = None

        abspath = self._abspath(path)
        os.makedirs(os.path.dirname(abspath), exist_ok=True)

        try:
            yield abspath
        except Exception as e:
            exception = e

        if obj is not None:
            # Update the original artefact with the changes
            self._updateArtefactObjects(obj)

        else:
            # Create a new artefact for the artefact
            self._loadArtefact(path)

        if exception:
            raise exception

class RemoteManager(Manager, RemoteInterface):
    """ Abstract Base Class for managers that will be working with remote artefacts so efficiency with fetching and
    pushing files is important for time and bandwidth
    """

    @staticmethod
    def _compare(dict1, dict2, key):
        # Extract the two sets of keys
        keys1, keys2 = set(dict1[key].keys()), set(dict2[key].keys())
        return keys1.difference(keys2), keys1.intersection(keys2), keys2.difference(keys1)

    @classmethod
    def _parseHierarchy(cls, path, _toplevel=None):

        # Store separately the directories and files of the path
        directories = {}
        files = {}

        # For each item process their checksums
        for item in os.listdir(path):

            # Identify their absolute path and relative manager path from the temporary local files
            abspath = os.path.join(path, item)

            if os.path.isdir(abspath):
                directories[abspath] = cls._parseHierarchy(abspath, _toplevel=path)

            else:
                files[abspath] = cls.md5(abspath)

        return {"directories": directories, "files": files}

    @classmethod
    def _compareHierarhy(cls, original, new):

        # Data containers for files and directory comparison
        toPush, toDelete = set(), set()

        # Compare the directories
        removed, editted, added = cls._compare(original, new, "directories")
        for directory in editted:
            put, delete = cls._compareHierarhy(original['directories'][directory], new['directories'][directory])

            # Union the result of the comparison on the sub directory level
            added |= put
            removed |= delete

        toPush |= added
        toDelete |= removed

        # Compare the files
        removed, editted, added = cls._compare(original, new, "files")
        for file in editted:
            if original['files'][file] != new['files'][file]:
                # The checksum of the files are not the same, therefore, the file has been editted and needs to be pushed
                added.add(file)

        toPush |= added
        toDelete |= removed

        return toPush, toDelete

    @contextlib.contextmanager
    def localise(self, artefact):
        obj, path = self._artefactFormStandardise(artefact)
        exception = None

        with tempfile.TemporaryDirectory() as directory:

            # Generate a temporay path for the file to be downloaded into
            local_path = os.path.join(directory, self.basename(path))

            # Get the contents and put it into the temporay directory
            if obj is not None:
                self.get(path, local_path)

                if os.path.isdir(local_path):
                    # To collected item is a directory - walk the directory and record its state
                    checksum = self._parseHierarchy(local_path)

                else:
                    # Generate a checksum for the file
                    checksum = self.md5(local_path)

            else:
                # No checksum for no object
                checksum = None

            # Return the local path to the object
            try:
                yield local_path
            except Exception as e:
                exception = e

            # The user has stopped interacting with the artefact - resolve any differences with manager
            if checksum:
                if os.path.isdir(local_path):
                    # Compare the new hiearchy - update only affected files/directories
                    put, delete = self._compareHierarhy(checksum, self._parseHierarchy(local_path))

                    # Define the method for converting the abspath back to the manager relative path
                    contexualise = lambda x: self.join(path, x[len(local_path)+1:], separator='/')

                    # Put/delete the affected artefacts
                    for abspath in put: self.put(abspath, contexualise(abspath))
                    for abspath in delete: self.rm(contexualise(abspath), recursive=True)

                elif self.md5(local_path) != checksum:
                    # The file has been changed - upload the file's contents
                    self.put(local_path, path)

            else:
                # New item - put the artefact into the manager
                self.put(local_path, path)

        if exception:
            raise exception



import collections
import typing
import urllib
import functools
import pkg_resources

MANAGERS = {}
INITIALISED_MANAGERS = {}

def initCache(function):
    """ Cache results and return previously created manager objects
    """

    functools.wraps(function)
    def wrapper(manager, **kwargs):

        identifier = hash((manager, "-".join(["{}-{}".format(k,v) for k,v in sorted(kwargs.items(), key=lambda x: x[0])])))

        if identifier in INITIALISED_MANAGERS:
            return INITIALISED_MANAGERS[identifier]

        manager = function(manager, **kwargs)

        INITIALISED_MANAGERS[identifier] = manager

        return manager

    return wrapper


def find(manager: str):
    """ Fetch the `Manager` class hosted on the 'stow_managers' entrypoint with the given name `manager` entry name.

    Args:
        manager: The name of the `Manager` class to be returned

    Returns:
        Manager: The `Manager` class for the manager name provided

    Raises:
        ValueError: In the event that a manager with the provided name couldn't be found
    """

    # Get the manager class for the manager type given - load the manager type if not already loaded
    lmanager = manager.lower()

    if lmanager in MANAGERS:
        mClass = MANAGERS[lmanager]

    else:
        foundManagerNames = []

        for entry_point in pkg_resources.iter_entry_points('stow_managers'):

            foundManagerNames.append(entry_point.name)

            if entry_point.name == lmanager:
                mClass = MANAGERS[lmanager] = entry_point.load()
                break

        else:
            raise ValueError("Couldn't find a manager called '{}' - found {} managers: {}".format(
                    manager,
                    len(foundManagerNames),
                    foundManagerNames
                )
            )

    return mClass

@initCache
def connect(manager: str, *, submanager: str = None, **kwargs):
    """ Find and connect to a `Manager` using its name (entrypoint name) and return an instance of that `Manager`
    initialised with the kwargs provided. A path can be provided as the location on the manager for a sub manager to be
    created which will be returned instead.

    Args:
        manager: The name of the manager class
        submanager: A path on the manager where a submanager is to be created
        kwargs: Keyworded arguments to be passed to the Manager init

    Returns:
        Manager: A storage manager or sub manager which can be used to put and get artefacts

    Note:
        References to `Manager` created by this method are stored to avoid multiple definitions of managers on similar
        locations.

        The stateless interface uses this method as the backend for its functions and as such you can fetch any active
        session by using this function rather than initalising a `Manager` directly
    """

    if submanager is not None:
        # Create the initial manager and return a sub-manager
        return connect(manager, **kwargs).submanager(submanager)

    # Find the class for the manager
    mClass = find(manager)

    # Create the Manager - pass all the kwarg arguments
    return mClass(**kwargs)

# Parsed URL tuple definition
ParsedURL = collections.namedtuple("ParsedURL", ["manager", "relpath"])

def parseURL(stowURL: str) -> ParsedURL:
    """ Parse the passed stow URL and return a ParsedURL a named tuple of manager and relpath

    Example:
        manager, relpath = stow.parseURL("s3://example-bucket/path/to/file)

        result = stow.parseURL("s3://example-bucket/path/to/file)
        result.manager
        result.relpath

    Args:
        stowURL: The path to be parsed and manager identified

    Returns:
        typing.NamedTuple: Holding the manager and relative path of
    """

    # Parse the url provided
    parsedURL = urllib.parse.urlparse(stowURL)

    # Handle protocol managers vs local file system
    if parsedURL.scheme and parsedURL.netloc:
        manager = find(parsedURL.scheme)
        scheme = parsedURL.scheme

    else:
        manager = find("FS")
        scheme = "FS"

    # Get the signature for the manager from the url
    signature, relpath = manager._signatureFromURL(parsedURL)

    # Has to use connect otherwise it will just create lots and lots of new managers
    return ParsedURL(connect(scheme, **signature), relpath)


from ..utils import connect

class ManagerReloader:
    def __new__(self, config):
        # This will create a new manager if it doesn't exist of fetch the one globally created
        return connect(**config)

class ManagerSeralisable:

    def __reduce__(self):
        return (ManagerReloader, (self.toConfig(),))

    def toConfig(self):
        raise NotImplementedError("toConfig needs to be defined sufficiently to run stow.connect(**config)")



import os
import io
import typing
import urllib
import hashlib

from ..artefacts import Artefact, File, Directory


class ClassMethodManager:
    """ Class method namespace for the Manager
    """

    SEPARATORS = ['\\', '/']

    @staticmethod
    def _splitArtefactUnionForm(artefact: typing.Union[Artefact, str]) -> typing.Tuple[typing.Union[Artefact, None], str]:
        """ Take an artefact or a string and return in a strict format the object and string representation. This allows
        methods to accept both and resolve and ensure.

        Only the path is guaranteed, the artefact object will be None if it is not passed

        Args:
            artefact: Type unknown, artefact object or path

        Returns:
            artefact: An artefact object or None
            path: the path passed or pull from the artefact object

        """
        if isinstance(artefact, Artefact):
            return artefact, artefact.path

        return None, artefact

    @classmethod
    def abspath(cls, artefact: typing.Union[Artefact, str]) -> str:
        """ Return a normalized absolute version of the path or artefact given.

        Args:
            artefact: The path or object whose path is to be made absolute and returned

        Returns:
            str: the absolute path of the artefact provided

        Raises:
            ValueError: Cannot make a remote artefact object's path absolute
        """
        _, path = cls._splitArtefactUnionForm(artefact)
        return os.path.abspath(path)

    @classmethod
    def basename(cls, artefact: typing.Union[Artefact, str]) -> str:
        """ Return the base name of an artefact or path. This is the second element of the pair returned by passing path
        to the function `split()`.

        Args:
            artefact: The path or object whose path is to have its base name extracted

        Returns:
            str: the basename
        """
        _, path = cls._splitArtefactUnionForm(artefact)
        return os.path.basename(path)

    @classmethod
    def name(cls, artefact: typing.Union[Artefact, str]) -> str:
        """ Return the name of an artefact or path (basename without extension).

        Args:
            artefact: The path or object whose path is to have its base name extracted

        Returns:
            str: the name e.g. /hello/there.txt => there
        """
        _, path = cls._splitArtefactUnionForm(artefact)
        basename = os.path.basename(path)
        index = basename.rfind('.')
        if index != -1:
            return basename[:index]
        return basename

    @classmethod
    def extension(cls, artefact: typing.Union[Artefact, str]) -> str:
        """ Return the extension of an artefact or path.

        Args:
            artefact: The path or object whose path is to have its base name extracted

        Returns:
            str: the extension e.g. /hello/there.txt => txt
        """
        _, path = cls._splitArtefactUnionForm(artefact)
        basename = os.path.basename(path)
        index = basename.rfind('.')
        if index != -1:
            return basename[index+1:]
        return ''

    @classmethod
    def commonpath(cls, paths: typing.Iterable[typing.Union[Artefact, str]]) -> str:
        """ Return the longest common sub-path of each pathname in the sequence paths

        Examples:
            commonpath(["/foo/bar", "/foo/ban/pip"]) == "/foo"

        Args:
            paths: Artefact/paths who will have their paths comparied to find a common path

        Returns:
            str: A valid owning directory path that is the shared owning directory for all paths

        Raises:
            ValueError: If there is no crossover at all
        """
        return os.path.commonpath([cls._splitArtefactUnionForm(path)[1] for path in paths])

    @classmethod
    def commonprefix(cls, paths: typing.Iterable[typing.Union[Artefact, str]]) -> str:
        """ Return the longest common string literal for a collection of path/artefacts

        Examples:
            commonpath(["/foo/bar", "/foo/ban/pip"]) == "/foo/ba"

        Args:
            paths: Artefact/paths who will have their paths comparied to find a common path

        Returns:
            str: A string that all paths startwith (may be empty string)
        """
        return os.path.commonprefix([cls._splitArtefactUnionForm(path)[1] for path in paths])

    @classmethod
    def dirname(cls, artefact: typing.Union[Artefact, str]) -> str:
        """ Return the directory name of path or artefact. Preserve the protocol of the path if a protocol is given

        Args:
            artefact: The artefact or path whose directory path is to be returned

        Returns:
            str: The directory path for the holding directory of the artefact
        """
        obj, path = cls._splitArtefactUnionForm(artefact)

        if obj is not None or path.find(":") == -1:
            # Obj path or path within no protocol and therefore no need to parse
            return os.path.dirname(path)

        else:
            # Preserve protocol (if there is one) - dirname the path
            result = urllib.parse.urlparse(artefact)
            return urllib.parse.ParseResult(
                result.scheme,
                result.netloc,
                os.path.dirname(result.path),
                result.params,
                result.query,
                result.fragment
            ).geturl()

    @staticmethod
    def expanduser(path: str) -> str:
        """ On Unix and Windows, return the argument with an initial component of ~ or ~user replaced by that user’s
        home directory.

        On Unix, an initial ~ is replaced by the environment variable HOME if it is set; otherwise the current user’s
        home directory is looked up in the password directory through the built-in module pwd. An initial ~user is
        looked up directly in the password directory.

        On Windows, USERPROFILE will be used if set, otherwise a combination of HOMEPATH and HOMEDRIVE will be used.
        An initial ~user is handled by stripping the last directory component from the created user path derived above.

        If the expansion fails or if the path does not begin with a tilde, the path is returned unchanged.

        Args:
            path: the path which may contain a home variable indicator to be expanded

        Returns:
            str: A path with the home path factored in - if applicable
        """
        return os.path.expanduser(path)

    @staticmethod
    def expandvars(path: str):
        """ Return the argument with environment variables expanded. Substrings of the form $name or ${name} are
        replaced by the value of environment variable name. Malformed variable names and references to non-existing
        variables are left unchanged.

        On Windows, %name% expansions are supported in addition to $name and ${name}.

        Args:
            path: A path which might contain variables to be expanded

        Returns:
            str: A string with any environment variables added
        """
        return os.path.expandvars(path)

    @staticmethod
    def isabs(path: str) -> bool:
        """ Return True if path is an absolute pathname.
        On Unix, that means it begins with a slash,
        on Windows that it begins with a (back)slash after chopping off a potential drive letter.

        Args:
            path: the path to be checked for being absolute
        """
        return os.path.isabs(path)

    @classmethod
    def join(cls, *paths: typing.Iterable[str], separator=os.sep, joinAbsolutes: bool = False) -> str:
        """ Join one or more path components intelligently. The return value is the concatenation of path and any
        members of *paths with exactly one directory separator following each non-empty part except the last,
        meaning that the result will only end in a separator if the last part is empty. If a component is an absolute
        path, all previous components are thrown away and joining continues from the absolute path component.

        Protocols/drive letters are perserved in the event that an absolute is passed in.

        Args:
            *paths: segments of a path to be joined together
            separator: The character to be used to join the path segments
            joinAbsolutes: Whether to stick to normal behaviour continue from absolute paths or join them in series

        Returns:
            str: A joined path
        """
        if not paths:
            return ""

        parsedResult = None  # Store the network information while path is joined
        joined = ""  # Constructed path

        for segment in paths:
            if isinstance(segment, Artefact):
                # Convert artefacts to paths
                segment = segment.path

            # Identify and record the last full
            presult = urllib.parse.urlparse(segment)
            if presult.scheme:
                parsedResult = presult
                segment = presult.path

            if joined:
                # A path is in the midst of being created

                if any(segment.startswith(sep) for sep in cls.SEPARATORS):
                    if joinAbsolutes:
                        joined = joined.rstrip('\\/') + segment

                    else:
                        joined = segment

                else:
                    if any(joined.endswith(sep) for sep in cls.SEPARATORS):
                        joined += segment

                    else:
                        joined += separator + segment

            else:
                joined = segment

        # Add back in the protocol if given
        if parsedResult:
            return urllib.parse.ParseResult(
                parsedResult.scheme,
                parsedResult.netloc,
                joined,
                parsedResult.params,
                parsedResult.query,
                parsedResult.fragment
            ).geturl()

        return joined

    @staticmethod
    def normcase(path: str) -> str:
        """ Normalize the case of a pathname. On Windows, convert all characters in the pathname to lowercase, and also
        convert forward slashes to backward slashes. On other operating systems, return the path unchanged.

        Args:
            path: path to normalise

        Returns:
            str: the path normalised
        """
        return os.path.normcase(path)

    @staticmethod
    def normpath(path: str) -> str:
        """ Normalize a pathname by collapsing redundant separators and up-level references so that A//B, A/B/, A/./B
        and A/foo/../B all become A/B.

        Args:
            path: the path whose to be

        Returns:
            str: The path transformed
        """
        # Check that the url is for a remote manager
        url = urllib.parse.urlparse(path)
        if url.scheme and url.netloc:
            # URL with protocol
            return urllib.parse.ParseResult(
                url.scheme,
                url.netloc,
                os.path.normpath(url.path).replace("\\", "/"),
                url.params,
                url.query,
                url.fragment
            ).geturl()

        # Apply the normal path - method to the path
        return os.path.normpath(path)

    @classmethod
    def realpath(cls, path: str) -> str:
        """ Return the canonical path of the specified filename, eliminating any symbolic links encountered in the path
        (if they are supported by the operating system).

        Args:
            path: the path to have symbolic links corrected

        Returns:
            str: the path with the symbolic links corrected
        """
        return os.path.realpath(path)

    @classmethod
    def relpath(cls, path: str, start=os.curdir) -> str:
        """ Return a relative filepath to path either from the current directory or from an optional start directory

        Args:
            path: the path to be made relative
            start: the location to become relative to
        """
        return os.path.relpath(path, start)

    @classmethod
    def samefile(cls, artefact1: typing.Union[Artefact, str], artefact2: typing.Union[Artefact, str]) -> bool:
        """ Check if provided artefacts are represent the same data on disk

        Args:
            artefact1: An artefact object or path
            artefact2: An artefact object or path

        Returns:
            bool: True if the artefacts are the same
        """
        obj1, path1 = cls._splitArtefactUnionForm(artefact1)
        obj2, path2 = cls._splitArtefactUnionForm(artefact2)

        if obj1 is None or obj2 is None:
            return os.path.samefile(path1, path2)

        else:
            return obj1 is obj2

    @staticmethod
    def sameopenfile(handle1: io.IOBase, handle2: io.IOBase) -> bool:
        """ Return True if the file descriptors fp1 and fp2 refer to the same file.
        """
        return os.path.sameopenfile(handle1, handle2)

    @classmethod
    def samestat(cls, artefact1: typing.Union[Artefact, str], artefact2: typing.Union[Artefact, str]) -> bool:
        """ Check if provided artefacts are represent the same data on disk

        Args:
            artefact1: An artefact object or path
            artefact2: An artefact object or path

        Returns:
            bool: True if the artefacts are the same
        """
        _, path1 = cls._splitArtefactUnionForm(artefact1)
        _, path2 = cls._splitArtefactUnionForm(artefact2)

        return os.path.samestat(path1, path2)

    @classmethod
    def split(cls, artefact: typing.Union[Artefact, str]) -> typing.Tuple[str, str]:
        """ Split the pathname path into a pair, (head, tail) where tail is the last pathname component and head is
        everything leading up to that.

        Args:
            artefact: the artefact to be split

        Returns:
            (dirname, basename): the split parts of the artefact
        """

        _, path = cls._splitArtefactUnionForm(artefact)
        return (cls.dirname(path), cls.basename(path))

    @classmethod
    def splitdrive(cls, path: str) -> typing.Tuple[str, str]:
        """ Split the pathname path into a pair (drive, tail) where drive is either a mount point or the empty string.

        Args:
            path: the path whose mount point/drive is to be removed

        Returns:
            (drive, path): tuple with drive string separated from the path
        """

        return os.path.splitdrive(path)

    @classmethod
    def splitext(cls, artefact: typing.Union[Artefact, str]) -> typing.Tuple[str, str]:
        """ Split the pathname path into a pair (root, ext) such that root + ext == path, and ext is empty or begins
        with a period and contains at most one period.

        Args:
            artefact: the artefact to have the extension extracted

        Returns:
            (root, ext): The root path without the extension and the extension
        """
        _, path = cls._splitArtefactUnionForm(artefact)
        return os.path.splitext(path)

    @staticmethod
    def md5(path):
        """ TODO """
        hash_md5 = hashlib.md5()
        with open(path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)

        return hash_md5.hexdigest()


__version__ = "1.1.1"

from .artefacts import Artefact, File, Directory, SubFile, SubDirectory
from .manager import Manager, SubManager
from . import exceptions

from .stateless import (
    find,
    connect,
    parseURL,
    artefact,
    abspath,
    basename,
    name,
    extension,
    commonpath,
    commonprefix,
    dirname,
    expanduser,
    expandvars,
    isabs,
    join,
    normcase,
    normpath,
    realpath,
    relpath,
    samefile,
    sameopenfile,
    samestat,
    split,
    splitdrive,
    splitext,
    md5,
    isfile,
    isdir,
    islink,
    ismount,
    getctime,
    getmtime,
    getatime,
    exists,
    lexists,
    touch,
    mkdir,
    localise,
    open,
    ls,
    get,
    put,
    cp,
    mv,
    sync,
    rm,
    supports_unicode_filenames,
)



class ArtefactNotFound(FileNotFoundError):
    """ An artefact cannot be found at a location """
    pass

class ArtefactNotMember(Exception):
    """ An artefact is not a member of a `Manager` or `Directory` """
    pass

class ArtefactTypeError(TypeError):
    """ Expected an artefact of another type """
    pass

class ArtefactNoLongerExists(OSError):
    """ The artefact cannot be accessed because its persistent representation has been deleted or moved. """
    pass

class OperationNotPermitted(Exception):
    """ You do not have permission to perform this action """
    pass

class InvalidPath(ValueError):
    """ Path given is not a valid stow path. See documentation """
    pass

class ArtefactNotAvailable(Exception):
    """ The artefact is not available to be able to fulfil the requested operation """
    pass


import os
from setuptools import setup, find_packages

# Read the index.md
with open(os.path.join(os.path.dirname(__file__), 'docs', 'index.md')) as fh:
    description = fh.read()

# Read in package requirements
with open(os.path.join(os.path.dirname(__file__), 'requirements.txt')) as fh:
    requires = fh.read().splitlines()

# Read in package test requirements
with open(os.path.join(os.path.dirname(__file__), 'test-requirements.txt')) as fh:
    testRequires = fh.read().splitlines()

def packageVersion(initpath: str) -> str:
    """ Get from the init of the source code the version string

    Params:
        initpath (str): path to the init file of the python package relative to the setup file

    Returns:
        str: The version string in the form 0.0.1
    """

    path = os.path.join(os.path.dirname(__file__), initpath)

    with open(path, "r") as handle:
        for line in handle.read().splitlines():
            if line.startswith("__version__"):
                return line.split("=")[1].strip().strip("\"'")

setup(
    name='stow',
    version=packageVersion("stow/__init__.py"),
    description="stow artefacts anywhere, with ease",
    long_description=description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
    ],

    url="https://github.com/Kieran-Bacon/stow",
    project_urls={
        "Documentation": "https://stow.readthedocs.io/en/latest/",
    },

    install_requires=requires,
    tests_require=testRequires,

    author="Kieran Bacon",
    author_email="kieran.bacon@outlook.com",

    packages=[package for package in find_packages() if not package.startswith("tests")],

    entry_points={
        'stow_managers': [
            'fs=stow.managers:FS',
            'lfs=stow.managers:FS',
            'aws=stow.managers:Amazon',
            's3=stow.managers:Amazon',
            'ssh=stow.managers:SSH',
        ]
    }
)


import abc
import io
import datetime
import contextlib
import typing
import weakref

from .class_interfaces import ManagerInterface, LocalInterface, RemoteInterface
from . import utils
from . import exceptions

class ArtefactReloader:
    def __new__(self, config, path):
        return utils.connect(**config)[path]

class Artefact:
    """ Artefacts are the items that are being stored - it is possible that through another mechanism that these items
    are deleted and they are no longer able to work

    Args:
        manager: The submanager this file belongs to
        path: The file's relative path
    """

    def __init__(
        self,
        manager: ManagerInterface,
        path: str
        ):

        self._manager = manager  # Link back to the owning manager
        self._path = path  # Relative path on manager
        self._exists = True # As you are created you are assumed to exist

    def __getattribute__(self, attr: str):

        if attr.startswith("_") or object.__getattribute__(self, '_exists'):
            return object.__getattribute__(self, attr)
        else:
            raise exceptions.ArtefactNoLongerExists(
                "Artefact {} {} no longer exists".format(type(self), object.__getattribute__(self, "_path"))
            )

    def __reduce__(self):
        return (ArtefactReloader, (self._manager.toConfig(), self._path))

    def __hash__(self): return hash(id(self))
    def __eq__(self, other): return hash(self) == hash(other)

    def __fspath__(self):
        if isinstance(self._manager, LocalInterface):
            # Get the filepath for the object - no protection for when the file is editted by os
            return self._manager._abspath(self._path)

        else:
            raise exceptions.ArtefactNotAvailable(
                "Cannot get filesystem path for {} as it is a remote object - "
                "we suggest that you open a localise context for this object and use the absolute path returned"
            )

    @property
    def manager(self):
        """ Return the manager object this Artefact belongs to """
        return self._manager

    @property
    def directory(self):
        """ Directory object this artefact exists within """
        return self._manager[self._manager.dirname(self._path)]

    @property
    @abc.abstractmethod
    def modifiedTime(self):
        """ UTC localised datetime of time file last modified by a write/append method """
        pass

    @property
    @abc.abstractmethod
    def createdTime(self):
        """ UTC localised datetime of time file last modified by a write/append method """
        pass

    @property
    @abc.abstractmethod
    def accessedTime(self):
        """ UTC localised datetime of time file last modified by a write/append method """
        pass

    @property
    def path(self):
        """ Return the manager relative path to this Artefact """
        return self._path
    @path.setter
    def path(self, path: str):
        """ Move the file on the target (perform the rename) - if it fails do not change the local file name """
        self._manager.mv(self, path)

    @property
    def basename(self):
        """ Basename of the artefact - holding directory path removed leaving filename and extension """
        return self._manager.basename(self.path)
    @basename.setter
    def basename(self, basename: str):
        self.manager.mv(self, self._manager.join(self._manager.dirname(self.path), basename, separator='/'))

    @property
    def name(self):
        """ Name of artefact - for `File` this is without extension """
        return self.basename
    @name.setter
    def name(self, name: str):
        self.basename = name

    @abc.abstractmethod
    @contextlib.contextmanager
    def localise(self):
        pass

    def save(self, path: str, force: bool = False):
        """ Save the artefact to a local location

        Args:
            path: A local path where the Artefact is to be saved
            force: Ignore artefacts at the destination location

        Raises:
            OperationNotPermitted: If the location given is a Directory and the get is not enforced
        """
        self._manager.get(self, path)

    def delete(self, force: bool = False):
        """ Delete this artefact from the disk

        Args:
            force: An "are you sure" for directories

        Raises:
            OperationNotPermitted: If directory and deletion has not been deleted
        """
        self.manager.rm(self, recursive=force)

class File(Artefact):
    """ A filesystem file object - a container of bytes representing some data

    Args:
        manager: The submanager this file belongs to
        path: The file's relative path
        modifiedTime: The time the file was last modified via a write/append operation
        size: The size in bytes of the file content
    """

    def __init__(
        self,
        manager,
        path: str,
        size: float,
        modifiedTime: datetime.datetime,
        createdTime: datetime.datetime = None,
        accessedTime: datetime.datetime = None
        ):
        super().__init__(manager, path)

        self._size = size  # The size in bytes of the object
        self._createdTime = createdTime  # Time the artefact was physically created
        self._modifiedTime = modifiedTime  # Time the artefact was last modified via the os
        self._accessedTime = accessedTime  # Time the artefact was last accessed

    def __len__(self): return self.size
    def __repr__(self):
        return '<stow.File: {} modified({}) size({} bytes)>'.format(self._path, self._modifiedTime, self._size)

    @property
    def name(self):
        if "." not in self.basename:
            return self.basename
        return self.basename[:self.basename.rindex(".")]

    @name.setter
    def name(self, name: str):
        ext = self.extension
        if ext:
            self.basename = "{}.{}".format(name, ext)

        else:
            self.basename = name

    @property
    def extension(self):
        """ File extension string - extention indicates file purpose and associated applications """
        if "." not in self.path:
            return ""
        return self.path[self.path.rindex(".")+1:]
    @extension.setter
    def extension(self, ext: str):
        self.basename = ".".join([self.name, ext])

    @property
    def content(self) -> bytes:
        """ file content as bytes """
        with self.open("rb") as handle:
            return handle.read()

    @content.setter
    def content(self, cont: bytes):
        if not isinstance(cont, bytes):
            raise ValueError("Cannot set the content of the file to non bytes type - {} given".format(type(cont)))

        with self.open("wb") as handle:
            handle.write(cont)

    @property
    def modifiedTime(self):
        """ UTC localised datetime of time file last modified by a write/append method """
        return self._modifiedTime

    @property
    def createdTime(self):
        """ UTC localised datetime of time file last modified by a write/append method """
        if self._createdTime is None:
            return self._modifiedTime
        return self._createdTime

    @property
    def accessedTime(self):
        """ UTC localised datetime of time file last modified by a write/append method """
        if self._accessedTime is None:
            return self._modifiedTime
        return self._accessedTime

    @property
    def size(self):
        """ Size of file content in bytes """
        return self._size

    @contextlib.contextmanager
    def localise(self) -> str:
        """ Localise this File artefact

        Returns:
            str: the absolute local path to the manager path
        """
        with self.manager.localise(self) as abspath:
            yield abspath

    @contextlib.contextmanager
    def open(self, mode: str = 'r', **kwargs) -> io.IOBase:
        """ Context manager to allow the pulling down and opening of a file """
        with self._manager.open(self, mode, **kwargs) as handle:
            yield handle

    def _update(self, other: Artefact):
        self._createdTime = other._createdTime
        self._modifiedTime = other._modifiedTime
        self._accessedTime = other._accessedTime
        self._size = other._size

class SubFile(File):
    """ A file object of a submanager. Wrapper for a complete Manager File

    Args:
        manager: The submanager this file belongs to
        path: The file's relative path
        file: The concrete file object this subfile wraps
    """

    def __init__(self, manager, path: str, file: File):
        super(File, self).__init__(manager, path)
        self._concrete = file

    def __len__(self): return len(self._concrete)
    def __repr__(self): return '<stow.SubFile for {}>'.format(self._concrete)

    @property
    def _modifiedTime(self): return self._concrete._modifiedTime
    @_modifiedTime.setter
    def _modifiedTime(self, time): self._concrete._modifiedTime = time

    @property
    def _createdTime(self): return self._concrete._createdTime
    @_createdTime.setter
    def _createdTime(self, time): self._concrete._createdTime = time

    @property
    def _accessedTime(self): return self._concrete._accessedTime
    @_accessedTime.setter
    def _accessedTime(self, time): self._concrete._accessedTime = time

    @property
    def content(self) -> bytes: return self._concrete.content
    @content.setter
    def content(self, content: bytes): self._concrete.content = content

    @property
    def _size(self): return self._concrete.size
    @_size.setter
    def _size(self, newSize): self._concrete._size = newSize

    @contextlib.contextmanager
    def open(self, mode: str = 'r', **kwargs) -> io.TextIOWrapper:
        """ Context manager to allow the pulling down and opening of a file """
        with self._concrete.open(mode, **kwargs) as handle:
            yield handle

    # def _update(self, other: Artefact): self._concrete._update(other)

class Directory(Artefact):
    """ A directory represents an local filesystems directory or folder. Directories hold references to other
    directories or files

    Args:
        manager (stow.Manager): The manager this directory object belongs to
        path (str): the manager relative path for the object
        contents (set): collection of artefacts which reside within this directory
        *,
        collected (bool): Toggle as to whether the directory contents has been collected (false when JIT Loading)
    """

    def __init__(
        self,
        manager,
        path: str,
        *,
        createdTime: datetime.datetime = None,
        modifiedTime: datetime.datetime = None,
        accessedTime: datetime.datetime = None
        ):
        super().__init__(manager, path)
        self._contents = weakref.WeakSet()
        self._collected = False

        self._createdTime = createdTime
        self._modifiedTime = modifiedTime
        self._accessedTime = accessedTime

    def __len__(self): return len(self.ls())
    def __iter__(self): return iter(self._contents)
    def __repr__(self): return '<stow.Directory: {}>'.format(self._path)
    def __contains__(self, artefact: typing.Union[Artefact, str]) -> bool:
        if isinstance(artefact, Artefact):
            return artefact.manager is self.manager and artefact in self._contents
        else:
            return self.manager.join(self._path, artefact, separator='/') in self.manager

    def _add(self, artefact: Artefact) -> None:
        assert isinstance(artefact, (File, Directory)) and not isinstance(artefact, (SubDirectory))
        self._contents.add(artefact)
    def _remove(self, artefact: Artefact) -> None: self._contents.remove(artefact)

    @property
    def createdTime(self):
        """ UTC localised datetime of time file last modified by a write/append method """
        if self._createdTime is None:
            if self._contents:
                return min(x.createdTime for x in self._contents)
            return None
        return self._createdTime

    @property
    def modifiedTime(self):
        """ UTC localised datetime of time file last modified by a write/append method """
        if self._modifiedTime is None:
            if self._contents:
                return max(x.modifiedTime for x in self._contents)
            return None
        return self._modifiedTime

    @property
    def accessedTime(self):
        """ UTC localised datetime of time file last modified by a write/append method """
        if self._accessedTime is None:
            if self._contents:
                return max(x.accessedTime for x in self._contents)
            return None
        return self._accessedTime

    def mkdir(self, path: str):
        """ Create a directory nested inside this `Directory` with the relative path given

        Args:
            path: Relative path to directory, path to new directory location

        Returns:
            Directory: The newly created directory object
        """
        return self.manager.mkdir(self.manager.join(self._path, path, separator='/', joinAbsolutes=True))

    def touch(self, path: str) -> File:
        """ Touch a file at given location relative to this Directory

        Args:
            path: The relative path to directory to touch new file

        Returns:
            File: The newly created file object
        """
        return self.manager.touch(self.manager.join(self._path, path, separator='/', joinAbsolutes=True))

    def relpath(self, artefact: typing.Union[Artefact, str]) -> str:
        """ Assuming the artefact is a member of this directory, return a filepath which is relative to this directory

        Args:
            artefact: the artefact who's path will be made relative

        Returns:
            str: the relative path to the artefact from this directory

        Raises:
            ArtefactNotMember: raised when artefact is not a member of the directory
        """

        # Get the path
        if isinstance(artefact, Artefact):
            path = artefact.path
        else:
            path = artefact

        # Raise error if the artefact is not a member of the directory
        if not path.startswith(self.path):
            raise exceptions.ArtefactNotMember(
                "Cannot create relative path for Artefact {} as its not a member of {}".format(artefact, self)
            )

        # Return the path
        return self.manager.relpath(path, self.path)

    @contextlib.contextmanager
    def localise(self, path: str = None) -> str:
        """ Localise an artefact this directory or a child artefact with the provided path.

        Args:
            path: Path of localisation

        Returns:
            str: the absolute local path to the manager path
        """
        with self.manager.localise(self if path is None else self.manager.join(self._path, path, separator='/')) as abspath:
            yield abspath

    @contextlib.contextmanager
    def open(self, path: str, mode: str = "r", **kwargs) -> io.IOBase:
        """ Open a file and create a stream to that file. Expose interface of `open`

        Args:
            path: Path to directory object
            mode: The open method
            kwargs: kwargs to be passed to the interface of open

        Yields:
            io.IOBase: An IO object depending on the mode for interacting with the file
        """
        with self.manager.open(self.manager.join(self._path, self.manager._managerPath(path)[1:], separator='/'), mode, **kwargs) as handle:
            yield handle

    def rm(self, path: str = None, recursive: bool = False):
        """ Remove an artefact at the given location

        Args:
            artefact: Path that is to be deleted
            recursive: If the target is a directory, whether to delete recursively the directories contents

        Raises:
            OperationNotPermitted: In the even the target is a directory and recursive has not been toggled
        """
        return self.manager.rm(self.manager.join(self.path, path, separator='/', joinAbsolutes=True), recursive)

    def _ls(self, recursive: bool = False):
        """ Get the current contents from this directory, do not update or edit state """

        if recursive:
            # Iterate through contents and recursively add lower level artifacts
            contents = set()
            for art in self._contents:
                if isinstance(art, Directory):
                    contents |= art._ls(True)

                contents.add(art)

            # Return all child content
            return contents

        # Make the content a set and return just this level
        return set(self._contents)

    def ls(self, path: str = None, recursive: bool = False) -> typing.Set[Artefact]:
        """ List the contents of this directory, or directory's directories.

        Args:
            path: The path to sub directory whose contents is to be returned
            recursive: Whether to recursively fetch all child contents for child directories

        Returns:
            typing.Set[Artefact]: The collection of objects within the targeted directory
        """
        return self._manager.ls(self.path if not path else self.manager.join(self.path, path, separator='/'), recursive=recursive)

    def isEmpty(self) -> bool:
        """ Check whether the directory has contents

        Returns:
            bool: True when there is at least one item in the directory False when the directory is empty
        """
        return not (bool(self._contents) or bool(len(self)))

    def empty(self):
        """ Empty the directory of contents """
        for artefact in self.ls():
            self._manager.rm(artefact, recursive=True)

    def _update(self, other: 'Directory'):

        self._contents = self._contents.union(other._contents)
        self._createdTime = other._createdTime
        self._modifiedTime = other._modifiedTime
        self._accessedTime = other._accessedTime
        self._collected = False

class SubDirectory(Directory):
    """ A directory object of a submanager. Wrapper for a complete Manager directory """

    def __init__(self, manager, path: str, directory: Directory):
        super(Directory, self).__init__(manager, path)  # Access the parent of Directory
        self._contents = weakref.WeakSet()
        self._collected = False
        self._concrete = directory

    @property
    def _collected(self): return self._concrete._collected
    @_collected.setter
    def _collected(self, value): pass


    def __len__(self): return len(self._concrete)
    def __repr__(self): return '<stow.SubDirectory for {}>'.format(self._concrete)
    def _add(self, artefact: Artefact) -> None:
        assert isinstance(artefact, (SubFile, SubDirectory))
        self._contents.add(artefact)

    @property
    def _modifiedTime(self): return self._concrete._modifiedTime

    @property
    def _createdTime(self): return self._concrete._createdTime

    @property
    def _accessedTime(self): return self._concrete._accessedTime

    def _update(self, other: Artefact):
        self._concrete._update(other._concrete)


import os
import datetime
import shutil
import pytz
import urllib

from ..artefacts import Artefact, File, Directory
from ..manager import LocalManager

class FS(LocalManager):
    """ Wrap a local filesystem (a networked drive or local directory)

    Args:
        path (str): The local relative path to where the manager is to be initialised
    """

    def __init__(self, path: str):
        # Record the local path to the original directory
        self._path = os.path.abspath(path)
        super().__init__()

    def __repr__(self): return '<Manager(FS): {}>'.format(self._path)

    def _abspath(self, managerPath: str) -> str:
        path = self.join(self._path, managerPath, joinAbsolutes=True)

        if os.name == 'nt':
            path = path.replace('/', '\\')

        return path

    def _identifyPath(self, managerPath: str):

        abspath = self._abspath(managerPath)

        if os.path.exists(abspath):

            stats = os.stat(abspath)

            # Created time
            createdTime = datetime.datetime.utcfromtimestamp(stats.st_ctime)
            createdTime = pytz.UTC.localize(createdTime)

            # Modified time
            modifiedTime = datetime.datetime.utcfromtimestamp(stats.st_mtime)
            modifiedTime = pytz.UTC.localize(modifiedTime)

            # Access time
            accessedTime = datetime.datetime.utcfromtimestamp(stats.st_atime)
            accessedTime = pytz.UTC.localize(accessedTime)

            if os.path.isfile(abspath):
                return File(
                    self,
                    managerPath,
                    stats.st_size,
                    modifiedTime,
                    createdTime,
                    accessedTime,
                )

            elif os.path.isdir(abspath):
                return Directory(
                    self,
                    managerPath,
                    createdTime=createdTime,
                    modifiedTime=modifiedTime,
                    accessedTime=accessedTime,
                )

        return None

    def _get(self, source: Artefact, destination: str):

        # Convert source path
        sourceAbspath = self._abspath(source.path)

        # Identify download method
        method = shutil.copytree if os.path.isdir(sourceAbspath) else shutil.copy

        # Download
        method(sourceAbspath, destination)

    def _getBytes(self, source: Artefact) -> bytes:

        with open(self._abspath(source.path), "rb") as handle:
            return handle.read()

    def _put(self, source: str, destination: str):

        # Convert destination path
        destinationAbspath = self._abspath(destination)

        # Ensure the destination
        os.makedirs(os.path.dirname(destinationAbspath), exist_ok=True)

        # Select the put method
        method = shutil.copytree if os.path.isdir(source) else shutil.copy

        # Perform the putting
        method(source, destinationAbspath)

    def _putBytes(self, fileBytes: bytes, destination: str):

        # Convert destination path
        destinationAbspath = self._abspath(destination)

        # Makesure the destination exists
        os.makedirs(os.path.dirname(destinationAbspath), exist_ok=True)

        # Write the byte file
        with open(destinationAbspath, "wb") as handle:
            handle.write(fileBytes)

    def _cp(self, source: Artefact, destination: str):
        self._put(self._abspath(source.path), destination)

    def _mv(self, source: Artefact, destination: str):

        # Convert the source and destination
        source, destination = self._abspath(source.path), self._abspath(destination)

        # Ensure the destination location
        os.makedirs(os.path.dirname(destination), exist_ok=True)

        # Move the source artefact
        os.rename(source, destination)

    def _ls(self, directory: str):

        # Get a path to the folder
        abspath = self._abspath(directory)

        # Iterate over the folder and identify every object - add the created
        for art in os.listdir(abspath):
            self._addArtefact(
                self._identifyPath(
                    self.join(directory, art, separator='/')
                )
            )

    def _rm(self, artefact: Artefact):

        # Convert the artefact
        artefact = self._abspath(artefact.path)

        # Select method for deleting
        method = shutil.rmtree if os.path.isdir(artefact) else os.remove

        # Remove the artefact
        method(artefact)

    @classmethod
    def _signatureFromURL(cls, url: urllib.parse.ParseResult):
        return {"path": "/"}, os.path.abspath(os.path.expanduser(url.path))

    def toConfig(self):
        return {'manager': 'FS', 'path': self._path}


import boto3
from botocore.exceptions import ClientError

import os
import re
import io
import typing
import urllib.parse
import enum

from ..artefacts import Artefact, File, Directory
from ..manager import RemoteManager
from .. import exceptions

class Amazon(RemoteManager):
    """ Connect to an amazon s3 bucket using an IAM user credentials or environment variables

    Params:
        bucket: The s3 bucket name
        aws_access_key_id (None): The access key for a IAM user that has permissions to the bucket
        aws_secret_access_key (None): The secret key for a IAM user that has permissions to the bucket
        region_name (None): The region of the user/bucket
        storage_class (STANDARD): The storage class type name e.g. STANDARD, REDUCED_REDUDANCY

    """

    # Define regex for the object key
    _LINE_SEP = "/"
    _S3_OBJECT_KEY = re.compile(r"^[a-zA-Z0-9!_.*'()-]+(/[a-zA-Z0-9!_.*'()-]+)*$")

    class StorageClass(enum.Enum):
        STANDARD = 'STANDARD'
        REDUCED_REDUNDANCY = 'REDUCED_REDUNDANCY'
        STANDARD_IA = 'STANDARD_IA'
        ONEZONE_IA = 'ONEZONE_IA'
        INTELLIGENT_TIERING = 'INTELLIGENT_TIERING'
        GLACIER = 'GLACIER'
        DEEP_ARCHIVE = 'DEEP_ARCHIVE'
        OUTPOSTS = 'OUTPOSTS'

    def __init__(
        self,
        bucket: str,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        region_name: str = None,
        storage_class: str = 'STANDARD'
    ):

        self._bucketName = bucket
        self._aws_access_key_id = aws_access_key_id
        self._aws_secret_access_key = aws_secret_access_key
        self._region_name = region_name
        self._storageClass = self.StorageClass(storage_class)

        self._s3Client = boto3.client(
            "s3",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name
        )
        self._s3Resource = boto3.resource(
            's3',
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name
        )

        # Create a paginator object for iterating through remote objects
        self._clientPaginator = self._s3Client.get_paginator('list_objects')

        # Create a reference to the AWS bucket - create a Directory to represent it
        self._bucket = self._s3Resource.Bucket(name=bucket) # pylint: disable=no-member

        super().__init__()

    def __repr__(self): return '<Manager(S3): {}>'.format(self._bucketName)

    def _abspath(self, managerPath: str) -> str:
        """ Difference between AWS and manager path is the removal of a leading '/'. As such remove the first character
        """
        abspath = managerPath.strip("/")
        assert not abspath or self._S3_OBJECT_KEY.match(abspath) is not None, "artefact name isn't accepted by S3: {}".format(abspath)
        return abspath

    def _identifyPath(self, managerPath: str) -> typing.Union[Artefact, None]:

        # Extract the key path
        key = self._abspath(managerPath)

        # Create a pagination that looks specifically at the manager path given
        pages = self._clientPaginator.paginate(
            Bucket=self._bucketName,
            Prefix=key,
            Delimiter="/"
        )

        # Iterate over the pages -
        for page in pages:
            if page is None:
                break

            # Check To see if there are any files with the given name
            if "Contents" in page:
                # There are files that lead with the manager path given
                for file in page["Contents"]:
                    if file["Key"] == key:
                        # We've found the file artefact that directly matches - create a file and return
                        return File(
                            self,
                            managerPath,
                            modifiedTime=file["LastModified"],
                            size=file["Size"]
                        )

            if "CommonPrefixes" in page:

                # Directories end in "/"
                key += "/"

                for directory in page["CommonPrefixes"]:
                    if directory["Prefix"] == key:
                        # The path is to the directory
                        return Directory(self, managerPath)

        return None

    def _loadArtefact(self, managerPath: str) -> Artefact:

        if managerPath in self._paths:
            # We may have already pulled the manager
            return super()._loadArtefact(managerPath)

        # Ensure the owning directory and fetch the directory object
        try:
            directory = self._ensureDirectory(self.dirname(managerPath))

        except (exceptions.ArtefactNotFound, exceptions.ArtefactTypeError) as e:
            raise exceptions.ArtefactNotFound("Cannot locate artefact {}".format(managerPath)) from e

        # Add all artefacts of the directory into the manager - adding the target artefact at the same time
        self._ls(directory.path)
        directory._collected = True

        # Check now that the directory has been downloaded and added that the original target exists
        if managerPath in self._paths:
            return self._paths[managerPath]

        else:
            raise exceptions.ArtefactNotFound("Cannot locate artefact {}".format(managerPath))

    def _get(self, source: Artefact, destination: str):

        # Convert manager path to s3
        keyName = self._abspath(source.path)

        # If the source object is a directory
        if isinstance(source, Directory):

            # Loop through all objects in the bucket and create them locally
            for object in self._bucket.objects.filter(Prefix=keyName):

                # Get the objects relative path to the directory
                relativePath = self.relpath(object.key, keyName)

                # Create object absolute path
                path = os.path.join(destination, relativePath)

                # Ensure the directory for that object
                os.makedirs(os.path.dirname(path), exist_ok=True)

                # Download the file to that location
                self._bucket.download_file(object.key, path)

        else:
            self._bucket.download_file(keyName, destination)

    def _getBytes(self, source: Artefact) -> bytes:

        # Get buffer to recieve bytes
        bytes_buffer = io.BytesIO()

        # Fetch the file bytes and write them to the buffer
        self._s3Client.download_fileobj(Bucket=self._bucketName, Key=self._abspath(source.path), Fileobj=bytes_buffer)

        # Return the bytes stored in the buffer
        return bytes_buffer.getvalue()

    def _put(self, source: str, destination: str):

        destination = self._abspath(destination)

        if os.path.isdir(source):
            # A directory of items is to be uploaded - walk local directory and uploaded each file

            sourcePathLength = len(source) + 1

            for root, dirs, files in os.walk(source):

                dRoot = self.join(destination, root[sourcePathLength:], separator='/')

                if not (dirs or files):
                    # There are no sub-directories or files to be uploaded
                    placeholder_path = self.join(dRoot, self._PLACEHOLDER, separator='/')
                    self._bucket.put_object(Key=placeholder_path, Body=b'', StorageClass=self._storageClass.value)
                    continue

                # For each file at this point - construct their local absolute path and their relative remote path
                for file in files:
                    self._bucket.upload_file(
                        os.path.join(root, file),
                        self.join(dRoot, file, separator='/'),
                        ExtraArgs = {'StorageClass': self._storageClass.value}
                    )

        else:
            # Putting a file
            self._bucket.upload_file(source, destination, ExtraArgs = {'StorageClass': self._storageClass.value})

    def _putBytes(self, fileBytes: bytes, destination: str):
        self._bucket.put_object(Key=self._abspath(destination), Body=fileBytes, StorageClass=self._storageClass.value)

    def _cpFile(self, source, destination):
        self._bucket.Object(destination).copy_from(CopySource={'Bucket': self._bucketName, 'Key': source})

    def _cp(self, source: Artefact, destination: str):

        # Convert the paths to s3 paths
        sourcePath, destinationPath = self._abspath(source.path), self._abspath(destination)

        # Detemine how to handle the source
        if isinstance(source, Directory):
            # Source is director - loop through and copy each file object
            for obj in self._bucket.objects.filter(Prefix=sourcePath):
                self._cpFile(obj.key, self.join(destinationPath, self.relpath(obj.key, sourcePath), separator='/'))

        else:
            # Source is a file - copy directly to location
            self._cpFile(sourcePath, destinationPath)

    def _mv(self, source: Artefact, destination: str):

        # Convert the paths to s3 paths
        sourcePath, destinationPath = self._abspath(source.path), self._abspath(destination)

        # Detemine how to handle the source
        if isinstance(source, Directory):
            # Source is director - loop through and copy each file object
            for obj in self._bucket.objects.filter(Prefix=sourcePath):
                self._cpFile(obj.key, self.join(destinationPath, self.relpath(obj.key, sourcePath), separator='/'))
                obj.delete()

        else:
            # Source is a file - copy directly to location
            self._cpFile(sourcePath, destinationPath)
            self._bucket.Object(sourcePath).delete()

    def _ls(self, directory: str):

        if directory != "/":
            key = self._abspath(directory) + "/"

        else:
            key = ""

        # Create a pagination that looks specifically at the manager path given
        pages = self._clientPaginator.paginate(
            Bucket=self._bucketName,
            Prefix=key,
            Delimiter="/"
        )

        # Iterate over the pages
        for page in pages:
            if page is None:
                break

            # Check To see if there are any files with the given name
            if "Contents" in page:
                # There are files that lead with the manager path given
                for file in page["Contents"]:

                    if self.basename(file["Key"]) == self._PLACEHOLDER:
                        # Don't list placeholders
                        continue

                    self._addArtefact(
                        File(
                            self,
                            "/" + file["Key"],
                            modifiedTime=file["LastModified"],
                            size=file["Size"]
                        )
                    )

            if "CommonPrefixes" in page:
                for directory in page["CommonPrefixes"]:
                    self._addArtefact(Directory(self, "/" + directory["Prefix"][:-1]))

    def _rm(self, artefact: Artefact):

        key = self._abspath(artefact.path)

        if isinstance(artefact, Directory):
            for obj in self._bucket.objects.filter(Prefix=key):
                obj.delete()

        else:
            self._bucket.Object(key).delete()

    @classmethod
    def _signatureFromURL(cls, url: urllib.parse.ParseResult):

        # Extract the query data passed into
        queryData = urllib.parse.parse_qs(url.query)

        signature = {
            "bucket": url.netloc,
            "aws_access_key_id": queryData.get("aws_access_key_id", [None])[0],
            "aws_secret_access_key": queryData.get("aws_secret_access_key", [None])[0],
            "region_name": queryData.get("region_name", [None])[0],
            "storage_class": queryData.get("storage_class", ['STANDARD'])[0],
        }

        return signature, (url.path or '/')

    def toConfig(self):
        return {
            'manager': 'AWS',
            'bucket': self._bucketName,
            'aws_access_key_id': self._aws_access_key_id,
            'aws_secret_access_key': self._aws_secret_access_key,
            'region_name': self._region_name,
            'storage_class': self._storageClass.value
        }
